import itertools
import numpy as np

from ...exceptions import InvalidScriptCharacter, InvalidScript, IncompatiblesScriptsLayers, TooManySingularSequences
from ...commons import TreeStructure
from ...constants import MAX_LAYER, MAX_SINGULAR_SEQUENCES, MAX_SIZE_HEADER, LAYER_MARKS, PRIMITIVES, \
    remarkable_multiplication_lookup_table, REMARKABLE_ADDITION, character_value, AUXILIARY_CLASS, VERB_CLASS, \
    NOUN_CLASS


class Script(TreeStructure):
    """ A parser is defined by a character (PRIMITIVES, REMARKABLE_ADDITION OR REMARKABLE_MULTIPLICATION)
     or a list of parser children. All the element in the children list must be an AdditiveScript or
     a MultiplicativeScript."""
    def __init__(self, children=None, character=None):
        super().__init__()

        if children:
            self.children = children
        else:
            self.children = []

        if character:
            self.character = character
        else:
            self.character = None

        # Layer of this parser
        self.layer = None

        # If it is a a paradigm
        self.paradigm = None

        # If the parser is composed with E
        self.empty = None

        # The number of singular sequence (if paradigm it is one, self)
        self.cardinal = None

        # The singular sequences
        self._singular_sequences = None
        self._singular_sequences_set = None

        # The contained paradigms (tables)
        self._tables = None
        self._cells = None
        self._tables_script = None

        # The canonical string to compare same layer and cardinal parser (__lt__)
        self.canonical = None

        # class of the parser, one of the following : VERB (1), AUXILIARY (0), and NOUN (2)
        self.script_class = None

    def __new__(cls, *args, **kwargs):
        """
        Need this to pickle scripts, the pickler use __hash__ method before unpickling the
        object attribute. Then need to pass the _str.
        """
        instance = super(Script, cls).__new__(cls)
        if 'str' in kwargs:
            instance._str = kwargs['str']

        return instance

    def __getnewargs_ex__(self):
        return ((), {
            'str': str(self)
        })

    def __add__(self, other):
        if not isinstance(other, Script):
            raise InvalidScript()

        return AdditiveScript(children=[self, other])

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def __hash__(self):
        """Since the IEML string for any proposition AST is supposed to be unique, it can be used as a hash"""
        return self._str.__hash__()

    def __lt__(self, other):
        if not isinstance(self, Script) or not isinstance(other, Script):
            return NotImplemented

        if self == other:
            return False

        if self.layer != other.layer:
            # order by layer
            return self.layer < other.layer
        else:
            # Cardinal of 1 the minimum, null_script is minimum
            if isinstance(self, NullScript):
                return True

            if isinstance(other, NullScript):
                return False

            if self.cardinal != other.cardinal:
                # then by number of singular sequence
                return self.cardinal < other.cardinal
            else:
                # Compare the canonical form
                if self.canonical != other.canonical:
                    return self.canonical < other.canonical
                elif self.layer != 0:
                    # layer != 0 => children is set, no children are for layer 0 (MultiplicativeScript)
                    if isinstance(self, other.__class__):
                        # if they are the same class
                        # Compare the children in alphabetical order
                        iterator = iter(other.children)
                        for s in self.children:
                            try:
                                o = iterator.__next__()
                                if o != s:
                                    return s < o
                            except StopIteration:
                                return False
                        # self have fewer elements, all equals to the first of other's children, self is lower.
                        return True
                    else:
                        # not an instance, one is multiplicative, other one is additive
                        # They have the same number of singular sequence so the multiplicative is fewer :
                        # each variable of the addition have less singular sequence than the multiplication parser
                        return isinstance(self, MultiplicativeScript)
                else:
                    # Layer 0
                    # Compare the value of the character or the sum
                    if isinstance(self, AdditiveScript):
                        # The character value is the sum of all character of the addition
                        self_char_value = sum((character_value[c.character] for c in self.children))
                    else:
                        self_char_value = character_value[self.character]

                    if isinstance(other, AdditiveScript):
                        # The character value is the sum of all character of the addition
                        other_char_value = sum((character_value[c.character] for c in other.children))
                    else:
                        other_char_value = character_value[other.character]

                    return self_char_value < other_char_value

    # def __getitem__(self, index):
    #     return self.children[index]

    def __contains__(self, item):
        if not isinstance(item, Script):
            return False

        if item.layer != self.layer:
            return False

        return item.singular_sequences_set.issubset(self.singular_sequences_set)

    def __len__(self):
        return self.cardinal

    def _build_tables(self):
        if self.cardinal == 1:
            self._cells = [np.array([[[self]]])]
            self._tables_script = [self]
        else:
            self._cells, self._tables_script = self._compute_cells()

    @property
    def cells(self):
        if self._cells is None:
            self._build_tables()
        return self._cells

    @property
    def tables_script(self):
        if self._tables_script is None:
            self._build_tables()
        return self._tables_script


    @property
    def singular_sequences(self):
        if self._singular_sequences is None:
            self._singular_sequences = self._compute_singular_sequences()

        return self._singular_sequences

    @property
    def singular_sequences_set(self):
        if self._singular_sequences_set is None:
            self._singular_sequences_set = set(self.singular_sequences)

        return self._singular_sequences_set

    def _compute_cells(self):
        pass

    def _compute_singular_sequences(self):
        pass


class AdditiveScript(Script):
    """ Represent an addition of same layer scripts."""
    def __init__(self, children=None, character=None):
        _character = None
        _children = []

        if children:
            _children = children

        if character in REMARKABLE_ADDITION:
            _character = character
            _children = REMARKABLE_ADDITION_SCRIPT[_character]
        elif character is not None:
            raise InvalidScriptCharacter(character)
        else:
            if len(_children) == 0:
                raise InvalidScript()

        # check all children are the same layer
        l = _children[0].layer
        for c in _children:
            if c.layer != l:
                raise IncompatiblesScriptsLayers(_children[0], c)

        if _children:

            to_remove = []
            to_add = []
            # remove the sub addition
            for c in _children:
                if isinstance(c, AdditiveScript):
                    to_remove.append(c)
                    to_add.extend(c.children)

            _children.extend(to_add)
            # Remove duplicate children
            _children = list(set(c for c in _children if c not in to_remove))

        # make a character with the children if possible
        if l == 0:
            _char_set = set(map(lambda e: str(e)[0], _children))
            for key, value in REMARKABLE_ADDITION.items():
                if _char_set == value:
                    _character = key
                    break

        super().__init__(children=_children, character=_character)

        if self.character:  # remarkable addition
            self.layer = 0
            self.empty = False
            self.paradigm = True
            self.cardinal = len(REMARKABLE_ADDITION[self.character])
        else:
            self.layer = self.children[0].layer
            self.empty = all((e.empty for e in self.children))
            self.paradigm = len(self.children) > 1 or any(child.paradigm for child in self.children)
            self.cardinal = sum((e.cardinal for e in self.children))

        if self.cardinal > MAX_SINGULAR_SEQUENCES:
            raise TooManySingularSequences(self.cardinal)

        self.script_class = max(c.script_class for c in self)

        self.__order()
        self._do_precompute_str()

    def _do_precompute_str(self):
        self._str = \
            (self.character + LAYER_MARKS[0]) if self.character is not None \
            else '+'.join([str(child) for child in self.children])

    def __order(self):
        # Ordering of the children
        self.children.sort()

        if self.layer == 0:
            value = 0b0
            for child in self:
                value |= character_value[child.character]
            self.canonical = bytes([value])
        else:
            self.canonical = b''.join([child.canonical for child in self])

    def _compute_singular_sequences(self):
        # Generating the singular sequence
        if not self.paradigm:
            return [self]
        else:
            # additive proposition has always children set
            s = [sequence for child in self.children for sequence in child.singular_sequences]
            s.sort()
            return s

    def _compute_cells(self):
        # we generate one table per children, unless one children is a singular sequence.
        # if so, we generate one column instead

        if any(not c.paradigm for c in self.children):
            # layer 0 -> column paradigm (like I: F: M: O:)
            return [np.array([[[s]] for s in self.singular_sequences])], [self]

        return [t for c in self.children for t in c.cells], [t for c in self.children for t in c.tables_script]


class MultiplicativeScript(Script):
    """ Represent a multiplication of three scripts of the same layer."""
    def __init__(self, substance=None, attribute=None, mode=None, children=None, character=None):
        if not (substance or children or character):
            raise InvalidScript()

        # Build children
        if children is None:
            children = [substance, attribute, mode]

        _children = []
        for child in children:
            if child is not None:
                _children.append(child)
            else:
                break

        # Replace all the corresponding children to character
        _character = None
        if character is not None:
            if character == 'E':
                raise InvalidScript()

            _character = character
            if _character in PRIMITIVES:
                _children = []
                layer = 0
            elif _character in REMARKABLE_MULTIPLICATION_SCRIPT:
                _children = REMARKABLE_MULTIPLICATION_SCRIPT[_character]
                layer = 1
            else:
                raise InvalidScriptCharacter(character)
        else:
            layer = _children[0].layer

        for i, c in enumerate(_children):
            elem = c
            if isinstance(c, AdditiveScript) and len(c.children) == 1:
                elem = c.children[0]
            _children[i] = elem

        # Replace the empty values
        for i, c in enumerate(_children):
            if c.empty:
                _children[i] = NullScript(layer=c.layer)

        # Fill the children to get a size of 3
        if _character not in PRIMITIVES:
            for i in range(len(_children), 3):
                _children.append(NullScript(layer=layer))

        # Add the character to children corresponding to specific combinaison
        _str_children = self._render_children(_children, _character)
        if _str_children in remarkable_multiplication_lookup_table:
            _character = remarkable_multiplication_lookup_table[_str_children]

        super().__init__(children=_children, character=_character)

        # Compute the attributes of this parser
        if self.character:
            self.layer = 0 if self.character in PRIMITIVES else 1
            self.paradigm = False
            self.cardinal = 1
            self.empty = self.character == 'E'
        else:
            self.layer = _children[0].layer + 1
            self.empty = all((e.empty for e in self.children))
            self.paradigm = any((e.paradigm for e in self.children))

            self.cardinal = 1
            for e in self.children:
                self.cardinal = self.cardinal * e.cardinal

        if self.layer == 0:
            self.script_class = VERB_CLASS if self.character in REMARKABLE_ADDITION['O'] else NOUN_CLASS
        else:
            self.script_class = self.children[0].script_class

        if self.layer != 0:
            # check number of children
            if not len(self.children) == 3:
                raise InvalidScript()

            # check every child of the same layer
            if not self.children[0].layer == self.children[1].layer == self.children[2].layer:
                raise InvalidScript()

            # check layer
            if not self.layer == self.children[0].layer + 1:
                raise InvalidScript()

        if self.cardinal > MAX_SINGULAR_SEQUENCES:
            raise TooManySingularSequences(self.cardinal)

        self.__order()
        self._do_precompute_str()

    def _render_children(self, children=None, character=None):
        if character:
            return character
        else:
            empty = True
            result = ''
            for i, c in enumerate(reversed(children)):
                if not c.empty:
                    empty = False

                if not empty or not c.empty or i == 2:
                    result = str(c) + result
            return result

    def _do_precompute_str(self):
        self._str = self._render_children(self.children, self.character) + LAYER_MARKS[self.layer]

    def __order(self):
        if self.layer == 0:
            self.canonical = bytes([character_value[self.character]])
        else:
            self.canonical = b''.join([child.canonical for child in self])

    def _compute_singular_sequences(self):
        # Generate the singular sequence
        if not self.paradigm:
            return [self]
        else:
            children_sequences = []
            for i in range(0, 3):
                if not self.children[i].empty:
                    children_sequences.append([(i, c) for c in self.children[i].singular_sequences])

            s = []
            for triplet in itertools.product(*children_sequences):
                children = self.children[:]
                for tpl in triplet:
                    children[tpl[0]] = tpl[1]

                sequence = MultiplicativeScript(children=children)
                s.append(sequence)

            s.sort()
            return s

    def _compute_cells(self):
        # check how many plurals child
        plurals_child = [(c, i) for i, c in enumerate(self.children) if c.cardinal != 1]

        if len(plurals_child) == 1:
            # only one plural child, we recurse
            v = plurals_child[0]

            # translate the child ss as ours
            map_seq = {s.children[v[1]]: s for s in self.singular_sequences}

            def resolve_ss(s):
                return map_seq[s]

            def map_script(s):
                return MultiplicativeScript(children=[
                    self.children[i] if i != v[1] else s for i in range(3)
                ])

            return [np.vectorize(resolve_ss)(c) for c in v[0].cells], [map_script(c) for c in v[0].tables_script]

        # more than one plural var, we build a multidimensional array
        # Check the table dimension
        if any(c.cardinal > MAX_SIZE_HEADER for c, _ in plurals_child):
            raise ValueError("The table defined by the script %s produce a table with more than %d headers."%
                             (str(self), MAX_SIZE_HEADER))

        # 1st dim the rows
        # 2nd dim the columns
        # 3rd dim the tabs
        result = np.zeros(shape=[plurals_child[i][0].cardinal if i < len(plurals_child)
                                 else 1 for i in range(3)], dtype=object)

        seq_index = [{s:i for i, s in enumerate(v[0].singular_sequences)} for v in plurals_child]

        for s in self.singular_sequences:
            res = [None, None, None]
            for i, v in enumerate(plurals_child):
                res[i] = seq_index[i][s.children[v[1]]]

            result[res[0], res[1], res[2]] = s

        if len(plurals_child) == 3:
            tables_script = [MultiplicativeScript(children=[self.children[0], self.children[1], ss])
                             for ss in self.children[2].singular_sequences]
        else:
            tables_script = [self]

        return [result], tables_script


class NullScript(Script):
    def __init__(self, layer):
        super().__init__(children=[])
        self.layer = layer
        self.paradigm = False
        self.empty = True
        self.cardinal = 1
        self.character = 'E'

        self._do_precompute_str()
        self.canonical = bytes([character_value[self.character]] * pow(3, self.layer))
        self.script_class = AUXILIARY_CLASS

    def __iter__(self):
        if self.layer == 0:
            return [].__iter__()

        return ([NULL_SCRIPTS[self.layer - 1]] * 3).__iter__()

    def _do_precompute_str(self):
        result = self.character
        for l in range(0, self.layer + 1):
            result = result + LAYER_MARKS[l]

        self._str = result

    def _compute_singular_sequences(self):
        return [self]


NULL_SCRIPTS = [NullScript(level) for level in range(0, MAX_LAYER)]

# Building the remarkable multiplication to parser
REMARKABLE_MULTIPLICATION_SCRIPT = {
    "wo": [MultiplicativeScript(character='U'), MultiplicativeScript(character='U'), NullScript(layer=0)],
    "wa": [MultiplicativeScript(character='U'), MultiplicativeScript(character='A'), NullScript(layer=0)],
    "y": [MultiplicativeScript(character='U'), MultiplicativeScript(character='S'), NullScript(layer=0)],
    "o": [MultiplicativeScript(character='U'), MultiplicativeScript(character='B'), NullScript(layer=0)],
    "e": [MultiplicativeScript(character='U'), MultiplicativeScript(character='T'), NullScript(layer=0)],

    "wu": [MultiplicativeScript(character='A'), MultiplicativeScript(character='U'), NullScript(layer=0)],
    "we": [MultiplicativeScript(character='A'), MultiplicativeScript(character='A'), NullScript(layer=0)],
    "u": [MultiplicativeScript(character='A'), MultiplicativeScript(character='S'), NullScript(layer=0)],
    "a": [MultiplicativeScript(character='A'), MultiplicativeScript(character='B'), NullScript(layer=0)],
    "i": [MultiplicativeScript(character='A'), MultiplicativeScript(character='T'), NullScript(layer=0)],

    "j": [MultiplicativeScript(character='S'), MultiplicativeScript(character='U'), NullScript(layer=0)],
    "g": [MultiplicativeScript(character='S'), MultiplicativeScript(character='A'), NullScript(layer=0)],
    "s": [MultiplicativeScript(character='S'), MultiplicativeScript(character='S'), NullScript(layer=0)],
    "b": [MultiplicativeScript(character='S'), MultiplicativeScript(character='B'), NullScript(layer=0)],
    "t": [MultiplicativeScript(character='S'), MultiplicativeScript(character='T'), NullScript(layer=0)],

    "h": [MultiplicativeScript(character='B'), MultiplicativeScript(character='U'), NullScript(layer=0)],
    "c": [MultiplicativeScript(character='B'), MultiplicativeScript(character='A'), NullScript(layer=0)],
    "k": [MultiplicativeScript(character='B'), MultiplicativeScript(character='S'), NullScript(layer=0)],
    "m": [MultiplicativeScript(character='B'), MultiplicativeScript(character='B'), NullScript(layer=0)],
    "n": [MultiplicativeScript(character='B'), MultiplicativeScript(character='T'), NullScript(layer=0)],

    "p": [MultiplicativeScript(character='T'), MultiplicativeScript(character='U'), NullScript(layer=0)],
    "x": [MultiplicativeScript(character='T'), MultiplicativeScript(character='A'), NullScript(layer=0)],
    "d": [MultiplicativeScript(character='T'), MultiplicativeScript(character='S'), NullScript(layer=0)],
    "f": [MultiplicativeScript(character='T'), MultiplicativeScript(character='B'), NullScript(layer=0)],
    "l": [MultiplicativeScript(character='T'), MultiplicativeScript(character='T'), NullScript(layer=0)]
}

# Building the remarkable addition to parser
REMARKABLE_ADDITION_SCRIPT = {key: [MultiplicativeScript(character=c) if c != 'E' else NullScript(layer=0) for c in REMARKABLE_ADDITION[key]] for key in REMARKABLE_ADDITION}




import ply.lex as lxr
import logging

logger = logging.getLogger(__name__)

tokens = (
   'PLUS',

    # Script specific
    'LAYER0_MARK',
    'LAYER1_MARK',
    'LAYER2_MARK',
    'LAYER3_MARK',
    'LAYER4_MARK',
    'LAYER5_MARK',
    'LAYER6_MARK',

    'PRIMITIVE',
    'REMARKABLE_ADDITION',
    'REMARKABLE_MULTIPLICATION'
)


def get_script_lexer(module=None):
    t_LAYER0_MARK = r'\:'
    t_LAYER1_MARK = r'\.'
    t_LAYER2_MARK = r'\-'
    t_LAYER3_MARK = r'[\'\’]'
    t_LAYER4_MARK = r'\,'
    t_LAYER5_MARK = r'\_'
    t_LAYER6_MARK = r'\;'

    t_PRIMITIVE = r'[EUASBT]'
    t_REMARKABLE_ADDITION = r'[OMFI]'
    t_REMARKABLE_MULTIPLICATION = r'wo|wa|y|o|e|wu|we|u|a|i|j|g|s|b|t|h|c|k|m|n|p|x|d|f|l'
    t_PLUS = r'\+'

    t_ignore = ' \t\n'

    # Error handling rule
    def t_error(t):
        logger.log(logging.ERROR, "Illegal character '%s'" % t.value[0])
        t.lexer.skip(1)

    return lxr.lex(module=module, errorlog=logging)




from bidict._bidict import bidict

COORDINATES_KINDS = ('t', 's', 'a', 'm', 'r', 'f')

RANKS = bidict({
    0: 'Term',
    1: 'Word',
    2: 'Sentence',
    3: 'SuperSentence',
    4: 'Text'
})

KIND_TO_RANK = {
    't': RANKS.inv['Text'],
    'r': RANKS.inv['Word'],
    'f': RANKS.inv['Word']
}


from ieml.commons import TreeStructure
from ieml.dictionary.dictionary import Dictionary
from ieml.syntax.commons import IEMLSyntax


class SyntaxTerm(IEMLSyntax):
    """
    Mapping {dictionary version -> Term}

    If multiple term are equivalent (under different dictionary versions), they share the same SyntaxTerm

    The instances of this class are determined by the last DictionaryVersion (ref. all the updates made by versions)
    """

    def __init__(self, term, literals=None):
        self.term = term
        self.dictionary_version = term.dictionary.version

        super().__init__(children=(), literals=literals)

    __hash__ = TreeStructure.__hash__

    def __str__(self):
        return self.term.__str__()

    def compute_str(self, children_str):
        return str(self)

    def set_dictionary_version(self, version):
        self._str = None
        self.term = Dictionary(version).translate_script_from_version(self.term.dictionary.version, self.term.script)

    def __getattr__(self, item):
        if item not in self.__dict__:
            return getattr(self.term, item)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.index == other.index

    def __gt__(self, other):
        return self.index > other.index




from functools import singledispatch

from ieml.exceptions import TermNotFoundInDictionary
from ieml.syntax.terms import SyntaxTerm
from .version import DictionaryVersion
from .script import Script
from .terms import Term
from .dictionary import Dictionary


def term(arg, dictionary=None):
    if not isinstance(dictionary, Dictionary):
        if isinstance(dictionary, (str, DictionaryVersion)):
            dictionary = Dictionary(dictionary)
        else:
            dictionary = Dictionary()

    try:
        return _term(arg, dictionary)
    except KeyError:
        raise TermNotFoundInDictionary(arg, dictionary)

@singledispatch
def _term(arg, dictionary):
    raise ValueError("Unsupported class %s for %s"%(arg.__class__.__name__, str(arg)))

_term.register(Term, lambda arg, dictionary: arg)
_term.register(int, lambda arg, dictionary: dictionary.index[arg])
_term.register(Script, lambda arg, dictionary: dictionary.terms[arg])


@_term.register(str)
def _term_str(arg, dictionary=None):
    if arg[0] == '[' and arg[-1] == ']':
        arg = arg[1:-1]


    return dictionary.terms[arg]


@_term.register(SyntaxTerm)
def _term_syntax(arg, dictionary):
    return arg.term


import logging

#
from ieml.calculation.distance import set_proximity_index
from ieml.filtering.filters import FilteringLevel, ParadigmaticProximityFilter, IndicatorFilter, BinaryFilter
from ieml.ieml_objects.hypertexts import Hypertext
from ieml.ieml_objects.sentences import SuperSentence, Sentence
from ieml.ieml_objects.terms import Term
from ieml.ieml_objects.texts import Text
from ieml.ieml_objects.words import Word

TOP_USL_TYPES = [SuperSentence, Hypertext, Text]


class USLSet:
    """A Wrapper object for a USL list"""

    def __init__(self, usl_list):
        self.usl_table = {}
        # splitting the USL list by type
        self._sort_usl_from_list(usl_list)

    def __len__(self):
        return sum([len(usl_level_list) for usl_level_list in self.usl_table.values()])

    @property
    def is_empty(self):
        """Returns true if there's at least one USL in the set"""
        for usl_list in self.usl_table.values():
            if usl_list:
                return False

        return True


    def _sort_usl_from_list(self, usl_list):
        """Sorting USL in a table, by USL filtering level"""
        for usl in usl_list:
            usl_type = FilteringLevel.get_usl_filtering_level(usl)
            if usl_type not in self.usl_table:
                self.usl_table[usl_type] = list()
            self.usl_table[usl_type].append(usl)

    def get_usls(self, usl_types=None):
        """Retrieve all USLs, or only USLs of a certain filtering type"""
        if usl_types is None:
            table_keys = self.usl_table.keys()
        else:
            table_keys = usl_types

        usls = []
        for usl_type in table_keys:
            if usl_type in self.usl_table:
                usls += self.usl_table[usl_type]

        return usls

    def set_usls(self, usl_list, usl_types=None):
        if usl_types is None:
            self.usl_table = {}
        else:
            for usl_t in usl_types:
                if usl_t in self.usl_table:
                    del self.usl_table[usl_t]
        self._sort_usl_from_list(usl_list)


class AbtractPipeline:

    def _filter(self, usl_set, query, final_pool_size, ratios_list=None):
        """Actual child class-dependent filtering function. Does the hard work"""
        pass

    def filter(self, usl_set, query, final_pool_size, ratios_list=None):
        """Does pre-checks on the pipeline's entry"""
        if usl_set.is_empty:
            logging.info("Empty USL set, aborting filtering")
            return usl_set
        else:
            return self._filter(usl_set, query, final_pool_size, ratios_list)

    def _check_ratios_list(self, filters_list, ratios_list):
        if len(ratios_list) == len(filters_list):
            self.chained_filters, self.filters_ratios = filters_list, ratios_list
            self.couples = list(zip(filters_list, ratios_list))
        else:
            raise Exception("Ratios count %i doesn't match the filter count %i" %
                            (len(ratios_list), len(filters_list)))

        if sum(ratios_list) != 1:
            raise Exception("Ratio sum doesn't sum to 1")

    @staticmethod
    def gen_pipeline_from_query(query_usl):
        return filtering_pipelines_mappings[FilteringLevel.get_usl_filtering_level(query_usl)]


class LinearPipeline(AbtractPipeline):
    """Pipeline for a simple linear F1 -> F2 -> ... schema for the filtering order.
    The Pipeline uses the defined ratios to reach the desired end ratio"""

    def __init__(self, filters_lists):
        self.filters_list = filters_lists
        self.filters_count = len(self.filters_list)
        self.scores_list , self.scores_dict= [],{}

    def log_appearance(self, start_pool_size, end_pool_size):
        logging.debug("Linear Pipeline:")
        filters_str_reprs = "-->".join("[%s, τ %1.3f]" % (str(filter), ratio) for filter, ratio in self.couples)
        logging.debug("[%i USL]-->%s-->[%i USL]" % (start_pool_size, filters_str_reprs, end_pool_size))

    def _store_scores(self, filter, score, filtered_usls):
        """For experimental uses: stores the scores of each USLs after filtering"""
        filtered_scores = { usl : score[usl] for usl in filtered_usls}
        self.scores_dict[filter] = filtered_scores
        self.scores_list.append((filter, filtered_scores))

    def _filter(self, usl_set, query_usl, final_pool_size, ratios_list=None):
        if ratios_list is not None:
            self._check_ratios_list(self.filters_list, ratios_list)
        else:
            self._check_ratios_list(self.filters_list, [1/self.filters_count] * self.filters_count)

        if not isinstance(usl_set, USLSet):
            # TODO : set the right exception for this case
            raise Exception("Expecting an USL set object, not an USL List")

        logging.debug("Starting filtering with the following pipeline: ")
        current_usl_pool = usl_set.get_usls()
        self.log_appearance(len(current_usl_pool), final_pool_size)
        master_ratio = final_pool_size / len(usl_set)

        for filter, ratio_power in self.couples:
            # if there are already not enough elements, let's not uselessly apply the filters
            if len(current_usl_pool) > final_pool_size:
                previous_count = len(current_usl_pool)
                current_usl_pool, scores = filter.filter(query_usl, current_usl_pool, pow(master_ratio, ratio_power))
                #self._store_scores(filter, scores, current_usl_pool)
                logging.debug("Removed %i elements at [%s] filter"
                              % (previous_count - len(current_usl_pool), str(filter)))
            else:
                break

        usl_set.set_usls(current_usl_pool)
        return usl_set


class ConditionalPipeline(AbtractPipeline):

    def __init__(self,filters_list, prefixed_filter):
        self.filters_list = filters_list
        self.prefixed_filter = prefixed_filter
        self.inner_pipeline = LinearPipeline(filters_list)

    @property
    def bf_lvl(self):
        return self.prefixed_filter.filtering_level

    def log_appearance(self, start_pool_size, end_pool_size):
        logging.debug("Conditional Pipeline:")
        prefilter_str_repr = "[%s]" % str(self.prefixed_filter)
        logging.debug("[%i USL]-->%s-->[UNKNOWN]-->[%i USL]" % (start_pool_size, prefilter_str_repr, end_pool_size))

    def _filter(self, usl_set, query, final_pool_size, ratios_list=None):
        self.log_appearance(len(usl_set.get_usls()), final_pool_size)

        higher_filtering_levels = FilteringLevel.get_higher_levels(self.bf_lvl)
        higher_level_usls = usl_set.get_usls(higher_filtering_levels)
        pre_filtered_usl_list = self.prefixed_filter.filter(query, higher_level_usls)
        output_list_count = len(pre_filtered_usl_list)
        logging.debug("Filtered %i of levels %s."
                      % (len(higher_level_usls) - output_list_count,
                         ", ".join([filt_lvl.name for filt_lvl in higher_filtering_levels])))
        if output_list_count == final_pool_size:
            logging.debug("Pretty lucky, ended up with the right number of USLs")
            return USLSet(pre_filtered_usl_list)

        elif output_list_count > final_pool_size:
            # linear pipeline for higher level USL
            logging.debug("Still %i too much. Picking a linear pipeline on the remaining USL equal or higher than %s"
                          % (output_list_count - final_pool_size,
                             str(self.prefixed_filter.filtering_level.name)))
            linear_pl = LinearPipeline(self.filters_list)
            return linear_pl.filter(USLSet(pre_filtered_usl_list),query, final_pool_size, ratios_list)

        else:

            if self.bf_lvl == FilteringLevel.MULTITERM_WORD:
                logging.debug("% i short of objective, picking a 2-filters linear pipeline for all remaining USLs"
                              % (final_pool_size - output_list_count))
                linear_pl = LinearPipeline(self.filters_list)
                usl_set.set_usls(pre_filtered_usl_list, higher_filtering_levels)
                return linear_pl.filter(usl_set, query, final_pool_size, ratios_list)
            else:
                # conditional pipeline on one level lower
                logging.debug("% i short of objective, picking a conditional pipeline for all remaining USLs"
                              % (final_pool_size - output_list_count))
                conditional_pl = filtering_pipelines_mappings[FilteringLevel(self.bf_lvl.value - 1)]
                usl_set.set_usls(pre_filtered_usl_list, higher_filtering_levels)
                if ratios_list is None:
                    return conditional_pl.filter(usl_set, query, final_pool_size, ratios_list)
                else:
                    # dividing the first ratio among the last ratios
                    new_ratio_list = [ ratio + (ratios_list[0] / len(ratios_list)) for ratio in ratios_list[1::]]
                    return conditional_pl.filter(usl_set, query, final_pool_size, new_ratio_list)


filtering_pipelines_mappings = {
    FilteringLevel.UNITERM_WORD: LinearPipeline([ParadigmaticProximityFilter(),
                                                 IndicatorFilter(set_proximity_index, level=Term)]),

    FilteringLevel.MULTITERM_WORD: ConditionalPipeline([IndicatorFilter(set_proximity_index, level=Term),
                                                        ParadigmaticProximityFilter()],
                                                       BinaryFilter(FilteringLevel.MULTITERM_WORD)),

    FilteringLevel.SENTENCE: ConditionalPipeline([IndicatorFilter(set_proximity_index, level=Word),
                                                  IndicatorFilter(set_proximity_index, level=Term),
                                                  ParadigmaticProximityFilter()],
                                                 BinaryFilter(FilteringLevel.SENTENCE)),

    FilteringLevel.SUPERSENTENCE: ConditionalPipeline([IndicatorFilter(set_proximity_index, level=Sentence),
                                                       IndicatorFilter(set_proximity_index, level=Word),
                                                       IndicatorFilter(set_proximity_index, level=Term),
                                                       ParadigmaticProximityFilter()],
                                                      BinaryFilter(FilteringLevel.SUPERSENTENCE))
}



from ieml.commons import TreeStructure
from ieml.exceptions import InvalidIEMLObjectArgument


class IEMLSyntaxType(type):
    """This metaclass enables the comparison of class times, such as (Sentence > Word) == True"""

    def __init__(cls, name, bases, dct):
        child_list = ['Morpheme', 'Word', 'Clause', 'Sentence',
                      'SuperClause', 'SuperSentence', 'Text',
                      'Hyperlink', 'Hypertext']

        if name in child_list:
            cls.__rank = child_list.index(name) + 1
        else:
            cls.__rank = 0

        super(IEMLSyntaxType, cls).__init__(name, bases, dct)

    def __hash__(self):
        return self.__rank

    def __eq__(self, other):
        if isinstance(other, IEMLSyntaxType):
            return self.__rank == other.__rank
        else:
            return False

    def __ne__(self, other):
        return not IEMLSyntaxType.__eq__(self, other)

    def __gt__(self, other):
        return IEMLSyntaxType.__ge__(self, other) and self != other

    def __le__(self, other):
        return IEMLSyntaxType.__lt__(self, other) and self == other

    def __ge__(self, other):
        return not IEMLSyntaxType.__lt__(self, other)

    def __lt__(self, other):
        return self.__rank < other.__rank

    def syntax_rank(self):
        return self.__rank


class IEMLSyntax(TreeStructure, metaclass=IEMLSyntaxType):
    closable = False

    def __init__(self, children, literals=None):
        super().__init__()
        self.children = tuple(children)

        if self.children:
            dictionary_version = self.children[0].dictionary_version
            if any(dictionary_version != c.dictionary_version for c in self.children[1:]):
                raise InvalidIEMLObjectArgument(self.__class__, "Multiple dictionary versions for this syntax object.")

            self.dictionary_version = dictionary_version
        elif not self.dictionary_version:
            raise InvalidIEMLObjectArgument(self.__class__, "No dictionary version specified for this syntax object.")


        _literals = []
        if literals is not None:
            if isinstance(literals, str):
                _literals = [literals]
            else:
                try:
                    _literals = tuple(literals)
                except TypeError:
                    raise InvalidIEMLObjectArgument(self.__class__, "The literals argument %s must be an iterable of "
                                                                    "str or a str."%str(literals))

        self.literals = tuple(_literals)
        self._do_precompute_str()

    def set_dictionary_version(self, version):
        self.dictionary_version = version
        for c in self.children:
            c.set_dictionary_version(version)

        self._str = None
        self._do_precompute_str()

    def __gt__(self, other):
        if not isinstance(other, IEMLSyntax):
            raise NotImplemented

        if self.__class__ != other.__class__:
            return self.__class__ > other.__class__

        return self._do_gt(other)

    def _do_gt(self, other):
        return self.children > other.children

    def compute_str(self, children_str):
        return '#'.join(children_str)

    def _compute_str(self):
        if self._str is not None:
            return self._str
        _literals = ''
        if self.literals:
            _literals = '<' + '><'.join(self.literals) + '>'

        return self.compute_str([e._compute_str() for e in self.children]) + _literals

    def _do_precompute_str(self):
        self._str = self._compute_str()

    def __getitem__(self, item):
        from ieml.usl.paths import Path, path, resolve

        if isinstance(item, str):
            item = path(item)

        if isinstance(item, Path):
            res = resolve(self, item)
            if len(res) == 1:
                return res.__iter__().__next__()
            else:
                return list(res)

        if isinstance(item, int):
            return self.children[item]

    def __len__(self):
        return len(self.children)



from .constants import MAX_SINGULAR_SEQUENCES


class InvalidPathException(Exception):
    def __init__(self, element, path, message):
        self.element = element
        self.path = path
        self.message = message

    def __str__(self):
        return "Can't access %s in %s, %s."%(str(self.path), str(self.element), str(self.message))


class CannotParse(Exception):
    def __init__(self, s, msg):
        self.s = s
        self.msg = msg

    def __str__(self):
        return "Unable to parse the following string '%s'. %s"%(str(self.s), str(self.msg))


class InvalidIEMLObjectArgument(Exception):
    def __init__(self, type, msg):
        self.type = type
        self.message = msg

    def __str__(self):
        return 'Invalid arguments to create a %s object. %s'%(self.type.__name__, str(self.message))


class InvalidTreeStructure(Exception):
    def __init__(self, msg):
        self.message = msg

    def __str__(self):
        return 'Invalid tree structure. %s'%str(self.message)

class CantGenerateElement(Exception):
    def __init__(self, msg):
        self.message = msg

    def __str__(self):
        return 'Unable to generate element. %s'%self.message


class InvalidScript(Exception):
    def __init__(self, msg=None):
        self.msg = msg

    def __str__(self):
        return "Invalid arguments to create a script. "+ self.msg if self.msg else ""


class InvalidScriptCharacter(InvalidScript):
    def __init__(self, character):
        self.character = character

    def __str__(self):
        return 'Invalid character %s for a parser.'%str(self.character)


class TooManySingularSequences(Exception):
    def __init__(self, num):
        self.num = num

    def __str__(self):
        return 'Too many singular sequences in the paradigms (%d, max %d).'%(self.num, MAX_SINGULAR_SEQUENCES)


class IncompatiblesScriptsLayers(InvalidScript):
    def __init__(self, s1, s2):
        self.s1 = s1
        self.s2 = s2

    def __str__(self):
        return 'Unable to add the two scripts %s, %s they have incompatible layers.'%(str(self.s1), str(self.s2))


class NoRemarkableSiblingForAdditiveScript(Exception):
    pass


class TermNotFoundInDictionary(Exception):
    def __init__(self, term, dictionary):
        self.message = "Cannot find term %s in the dictionary %s" % (str(term), str(dictionary.version))

    def __str__(self):
        return self.message


class ScriptNotDefinedInVersion(Exception):
    def __init__(self, script, version):
        self.message = "Script {1} not defined in the dictionary version {1}".format(str(script), str(version))

    def __str__(self):
        return self.message


from bidict import bidict


GRAMMATICAL_CLASS_NAMES = bidict({
    0: 'AUXILIARY',
    1: 'VERB',
    2: 'NOUN'
})

AUXILIARY_CLASS = GRAMMATICAL_CLASS_NAMES.inv['AUXILIARY']
VERB_CLASS = GRAMMATICAL_CLASS_NAMES.inv['VERB']
NOUN_CLASS = GRAMMATICAL_CLASS_NAMES.inv['NOUN']

LANGUAGES = {
    'fr',
    'en'
}

MORPHEME_SIZE_LIMIT = 6
MAX_NODES_IN_SENTENCE = 20
MAX_DEPTH_IN_HYPERTEXT = 8
MAX_NODES_IN_HYPERTEXT = 20
MAX_SINGULAR_SEQUENCES = 360
MAX_SIZE_HEADER = 12
MAX_LAYER = 6

# max number of terms (arbitrary, this value was chosen for allocate the node id range for drupal)
MAX_TERMS_IN_DICTIONARY = 18999

LAYER_MARKS = [
    ':',
    '.',
    '-',
    '\'',
    ',',
    '_',
    ';'
]

PRIMITIVES = {
    'E',
    'U',
    'A',
    'S',
    'B',
    'T'
}

character_value = {
    'E': 0x1,
    'U': 0x2,
    'A': 0x4,
    'S': 0x8,
    'B': 0x10,
    'T': 0x20

}

remarkable_multiplication_lookup_table = {
    "U:U:": "wo", "U:A:": "wa", "U:S:": "y", "U:B:": "o", "U:T:": "e",
    "A:U:": "wu", "A:A:": "we", "A:S:": "u", "A:B:": "a", "A:T:": "i",
    "S:U:": "j",  "S:A:": "g",  "S:S:": "s", "S:B:": "b", "S:T:": "t",
    "B:U:": "h",  "B:A:": "c",  "B:S:": "k", "B:B:": "m", "B:T:": "n",
    "T:U:": "p",  "T:A:": "x",  "T:S:": "d", "T:B:": "f", "T:T:": "l"
}

REMARKABLE_ADDITION = {
    "O": {'U', 'A'},
    "M": {'S', 'B', 'T'},
    "F": {'U', 'A', 'S', 'B', 'T'},
    "I": {'E', 'U', 'A', 'S', 'B', 'T'}
}




from _operator import mul
from collections import namedtuple, defaultdict
from itertools import chain, product

from functools import reduce

from ieml.syntax.terms import SyntaxTerm
from ...syntax import SuperSentence, Sentence, Text, Word
from ...dictionary import Term
from .constants import COORDINATES_KINDS
from .exceptions import PathError

Context = namedtuple("Context",
                     ['accept',   # the types this context can deference
                      'conserve', # if this context must imidiately swith to a lower type (mode and morphems)
                      'switch'])  # the mapping type -> type this context can switch

COORDINATES_CONTEXTS = {
    't': Context(accept={Text}, conserve=False, switch={
        Text: {SuperSentence, Sentence, Word, SyntaxTerm}
    }),
    'a': Context(accept={SuperSentence, Sentence}, conserve=True, switch={
        SuperSentence: {Sentence},
        Sentence: {Word}
    }),
    's': Context(accept={SuperSentence, Sentence}, conserve=True, switch={
        SuperSentence: {Sentence},
        Sentence: {Word}
    }),
    'm': Context(accept={SuperSentence, Sentence}, conserve=False, switch={
        SuperSentence: {Sentence},
        Sentence: {Word}
    }),
    'r': Context(accept={Word}, conserve=False, switch={
        Word: {SyntaxTerm}
    }),
    'f': Context(accept={Word}, conserve=False, switch={
        Word: {SyntaxTerm}
    })
}


class Path:
    # represent a path.
    # A path is an element of the form
    # this correspond to a descent in an ieml tree
    #
    # 't0:sa0:r0'
    # 't:s:(r0+f1)'
    # 't:(sm0+sa0:m0):r1'
    # '(t0+t1):s:s:f'
    #
    # A path have a context, the types of ieml-object that this path can deference.
    # The type of the ieml_object inferred by this path on a well typed ieml-object is determined by the context at the
    # end of the string.
    #
    # There are 5 different context, for each ieml-object type. A context switch occur with ':'.
    # - the Text context :
    #    't' is the only literal. If indexed like 'tn' where 'n' is an integer, it represent the n-th element in the
    # text object. This context allow only one character ('t' or 'tn') then it is always followed by ':' or nothing.
    # It can be followed by a super-sentence, a sentence or a word context.
    #
    # - the Super Sentence and Sentence ctx:
    # Begin with 's' or 'sn' but if the 'n' is an integer, it will be ignored. Can be followed by 'an' or 'mn'
    # where 'n' is an integer. When it is followed by 'an' it lead to the the n-th attribute of the current node. Same
    # for 'mn', but for the mode of the n-th attribute (the clause of this node, the n-th attribute and his mode). If
    # it is a mode, it can only be followed by a context switch ':'.
    # Supersentence lead to Sentence ctx and sentence to word ctx.
    #
    # - the Word context:
    # Only one literal of the form 'r' | 'rn' | 'f' | 'fn' where 'n' is an integer. These literal cannot be followed by
    # another character. If 'r' it lead all the root terms (the morphem), 'rn' lead only the n-th in the root morphem.
    # Same for 'f'|'fn' and the flexing morphem.
    #
    #  Ex: 't0:sa0:r' -> In a text context (or usl), the first element of the text, which is a sentence,
    #  then in a sentence context, the word of the first attribute of the root (index on 's' is ignored),
    #  then in a word context, the roots terms.
    #
    #
    #
    # the context of a path can be accessed:
    #   p.context.accept is a set of accepted types to deference
    #   p.context.switch is a dict of accepted types to inferred types i.e. the type of b = a[p] where a type is in
    # p.context.accept. len(p.context.switch) == len(p.context.accept)
    #
    # We add the '+' notation. An additive path is defined by the sum of multiple paths.
    # To get the list of sub path : p.develop
    #
    # p.cardinal == len(p.develop)
    #
    #


    def __init__(self, children):
        try:
            children = list(children)
        except TypeError:
            raise ValueError("The children must be iterable.")

        if any(not isinstance(child, Path) for child in children):
            raise ValueError('The children must be path instances.')

        # children unfold
        _children = []
        for c in children:
            if isinstance(c, self.__class__):
                _children += c.children
            else:
                if len(c.children) == 1:
                    _children.append(c.children[0])
                else:
                    _children.append(c)

        self.children = tuple(_children)
        self.cardinal = None
        self._development = None
        self.context = None

    @property
    def develop(self):
        if self._development is None:
            if self.cardinal == 1:
                self._development = (self,)
            else:
                self._development = self._develop()

        return self._development

    def _develop(self):
        raise NotImplemented

    def _resolve_context(self):
        if isinstance(self, Coordinate):
            self.context = COORDINATES_CONTEXTS[self.kind]
            return

        # the development is either a mul of coords or a context of mul and coords

        accept = set()
        switch = defaultdict(set)
        conserve = False

        for d in self.develop:
            if isinstance(d, Coordinate):
                accept |= d.context.accept
                conserve = conserve or d.context.conserve
                for c in d.context.accept:
                    switch[c] |= d.context.switch[c]

            elif isinstance(d, MultiplicativePath):
                # must ensure that the context is conserved until the last element
                for ctx in d.children[0].context.accept:
                    for c in d.children[:-1]:
                        # if the coord not conserve the current context
                        if not c.context.conserve:
                            raise PathError("Invalid developed path, context discontinuity.", d)

                        if ctx not in c.context.accept:
                            # the context is lost
                            break
                    else:
                        last = d.children[-1].context
                        # check the last coord if it accept the element
                        if ctx in last.accept:
                            # if the last element conserve the context, set it to true (at least one path conserve)
                            conserve = conserve or last.conserve

                            # add the passed type to the accept list
                            accept |= {ctx}

                            # add the switch of the last element
                            switch[ctx] |= last.switch[ctx]

            else:
                # instance of context path
                for str_acc in d.children[0].context.accept:
                    stack = {str_acc}
                    for c in d.children:
                        _stack = set()
                        for s in stack:
                            if s in c.context.accept:
                                _stack |= c.context.switch[s]

                        if not _stack:
                            break

                        stack = _stack
                    else:
                        accept |= {str_acc}
                        switch[str_acc] |= stack

        if not accept:
            raise PathError("No context match this path.", self)

        self.context = Context(accept=accept, conserve=conserve, switch=dict(switch))

    def __mul__(self, other):
        if isinstance(other, str):
            from .tools import path
            other = path(other)

        if isinstance(other, Path):
            return ContextPath([self, other])

        raise NotImplemented

    def __add__(self, other):
        if isinstance(other, str):
            from .tools import path
            other = path(other)

        if isinstance(other, Path):
            return AdditivePath([self, other])

        raise NotImplemented

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        return hash(self.__str__())


class ContextPath(Path):
    def __init__(self, children):
        if not children:
            raise ValueError('Must be a non empty children.')

        super().__init__(children)

        if len(self.children) < 2:
            raise ValueError("A context path must have at least two children.")

        self.cardinal = reduce(mul, [c.cardinal for c in self.children])
        self._resolve_context()

    def __str__(self):
        result = []
        for c in self.children:
            if isinstance(c, AdditivePath):
                result.append('(%s)'%str(c))
            else:
                result.append(str(c))

        return ':'.join(result)

    def _develop(self):
        return tuple(ContextPath(p) for p in product(*[c.develop for c in self.children]))


class AdditivePath(Path):
    def __init__(self, children):
        if not children:
            raise ValueError('Must be a non empty children.')

        super().__init__(children)

        self.cardinal = sum(c.cardinal for c in self.children)
        self._resolve_context()

    def __str__(self):
        return '+'.join(map(str, self.children))

    def _develop(self):
        return tuple(chain.from_iterable(c.develop for c in self.children))


class MultiplicativePath(Path):
    def __init__(self, children):
        if not children:
            raise ValueError('Must be a non empty children.')

        super().__init__(children)

        self.cardinal = reduce(mul, [c.cardinal for c in self.children])

        self._resolve_context()

    def __str__(self):
        result = ''
        for c in self.children:
            if isinstance(c, AdditivePath):
                result += '(%s)'%str(c)
            else:
                result += str(c)

        return result

    def _develop(self):
        return tuple(MultiplicativePath(p) for p in product(*[c.develop for c in self.children]))


class Coordinate(Path):
    def __init__(self, kind, index=None):

        if not isinstance(kind, str) or not len(kind) == 1 or not kind in COORDINATES_KINDS:
            raise ValueError("A coordinate kind must be one of the following (%s)."%
                             ', '.join(map(str, COORDINATES_KINDS)))

        self.kind = kind

        if index is not None:
            if not isinstance(index, int):
                raise ValueError("Coordinate index must be int.")

            self.index = index
        else:
            self.index = None

        # no children
        super().__init__(())

        self.cardinal = 1

        self._resolve_context()

    def __str__(self):
        return self.kind + (str(self.index) if self.index is not None else '')




from collections import defaultdict

from .commons import IEMLSyntax
from ..constants import MAX_NODES_IN_HYPERTEXT, MAX_DEPTH_IN_HYPERTEXT
from ..exceptions import InvalidIEMLObjectArgument, InvalidTreeStructure
from ..dictionary import Term
from .texts import Text
from .sentences import Sentence, SuperSentence
from .words import Word
from .tree_graph import TreeGraph


class PropositionPath:
    def __init__(self, list_proposition):
        super().__init__()
        self.path = tuple(list_proposition)

    def in_text(self, text):
        current = text
        for p in self.path:
            if not p in current.children:
                return False
            current = p
        return True

    def __hash__(self):
        return hash(self.path)

    def __eq__(self, other):
        if isinstance(other, PropositionPath):
            o = other.path
        else:
            o = tuple(other)

        return self.path == o

    @property
    def end(self):
        return self.path[-1]

    def __gt__(self, other):
        raise NotImplemented

    def __str__(self):
        return '/'.join((str(p) for p in self.path))

    def _compute_str(self):
        return '/'.join((str(p) for p in self.path))


class Hyperlink(IEMLSyntax):
    def __init__(self, substance=None, attribute=None, mode=None, children=None):

        if children:
            raise NotImplemented

        if not isinstance(substance, Text):
            raise InvalidIEMLObjectArgument(Hyperlink, "The substance %s must be a Text instance."%str(substance))

        if not isinstance(attribute, Text):
            raise InvalidIEMLObjectArgument(Hyperlink, "The attribute %s must be a Text instance."%str(substance))

        if substance == attribute:
            raise InvalidIEMLObjectArgument(Hyperlink, "The substance and the attribute %s must be distinct."%str(substance))

        if not isinstance(mode, PropositionPath):
            raise InvalidIEMLObjectArgument(Hyperlink, "The mode %s must be a PropositionPath instance."%str(substance))

        if not mode.in_text(substance):
            raise InvalidIEMLObjectArgument(Hyperlink, "The mode %s should be a PropositionPath in the text in"
                                                       " substance %s."%(str(mode), str(substance)))

        if not isinstance(mode.end, (Word, Sentence, SuperSentence)):
            raise InvalidIEMLObjectArgument(Hyperlink, "The mode %s must be a PropositionPath pointing to a Word, a "
                                                       "Sentence or a SuperSentence."%str(mode))

        super().__init__((substance, attribute, mode))

    @property
    def start(self):
        return self.children[0]

    @property
    def end(self):
        return self.children[1]

    @property
    def path(self):
        return self.children[2]

    def compute_str(self, children_str):
        return '(' + ','.join(children_str) + ')'


class Hypertext(IEMLSyntax):
    closable = True

    def __init__(self, children):

        try:
            _children = [e for e in children]
        except TypeError:
            raise InvalidIEMLObjectArgument(Hypertext, "The argument %s is not iterable." % str(children))

        if not all(isinstance(e, Hyperlink) for e in _children):
            raise InvalidIEMLObjectArgument(Hypertext, "The argument %s must be a list of Hyperlink instance."%
                                            str(_children))

        try:
            self.tree_graph = TreeGraph(((c.start, c.end, c) for c in _children))
        except InvalidTreeStructure as e:
            raise InvalidIEMLObjectArgument(Hypertext, e.message)

        if len(self.tree_graph.nodes) > MAX_NODES_IN_HYPERTEXT:
            raise InvalidIEMLObjectArgument(Hypertext, "Too many text in this hypertext: %d>%d."%
                                            (len(self.tree_graph.nodes), MAX_NODES_IN_HYPERTEXT))

        if len(self.tree_graph.stages) > MAX_DEPTH_IN_HYPERTEXT:
            raise InvalidIEMLObjectArgument(Hypertext, "The depth of the tree graph of this hypertext is too big: %d>%d"
                                            %(len(self.tree_graph.stages), MAX_DEPTH_IN_HYPERTEXT))

        #TODO sort the children
        super().__init__(_children)

    def compute_str(self, children_str):
        def render_text(text):
            hyperlinks = defaultdict(lambda: list())
            # e[1] is the hyperlink list starting at text
            for h in [e[1] for e in self.tree_graph.transitions[text]]:
                hyperlinks[h.path].append(h.end)

            def _render(path, current):

                if isinstance(current, Term):# or tuple(_path) not in hyperlinks:
                    return str(current)

                return current.compute_str([_render(path + [c], c) for c in current]) +\
                    ''.join(render_text(t) for t in hyperlinks[tuple(path)])

            return _render([], text)

        return render_text(self.tree_graph.root)



import ply.lex as lxr
import logging

tokens = (
   'IEML_OBJECT',
   # 'L_CURLY_BRACKET',
   # 'R_CURLY_BRACKET',
)


def get_lexer(module=None):
    t_IEML_OBJECT = r'.+'
    # t_L_CURLY_BRACKET = r'\{'
    # t_R_CURLY_BRACKET = r'\}'

    t_ignore  = ' \t\n'

    # Error handling rule
    def t_error(t):
        logger = logging.getLogger(__name__)
        logger.log(logging.ERROR, "Illegal character '%s'" % t.value[0])
        t.lexer.skip(1)

    return lxr.lex(module=module, errorlog=logging)



import itertools as it
import numpy as np
from bidict import bidict

from .script import MultiplicativeScript, Script, AdditiveScript, remarkable_multiplication_lookup_table


def old_canonical(script_ast):
    result = ''
    for byte in script_ast.canonical:
        result += chr(byte + ord('a') - 1)
    return [result]


def factor(sequences):
    layer = next(iter(sequences)).layer

    if layer == 0:
        return list(sequences)

    if len(sequences) == 1:
        return list(sequences)

    # holds the attributes/substances/modes as individual sets in primitives[0]/primitives[1]/primitives[2] respectively
    primitives = (set(seme) for seme in zip(*sequences))

    # same but now there is a bijection between the coordinate system and the primitives semes
    primitives = [bidict({i: s for i, s in enumerate(p_set)}) for p_set in primitives]

    # hold the mapping coordinate -> parser
    scripts = {tuple(primitives[i].inv[seme] for i, seme in enumerate(s)):s for s in sequences}

    # hold the primitive as coodinate described in scripts keys
    shape = tuple(len(p) for p in primitives)
    topology = np.full(shape, False, dtype=bool)
    for s in scripts:
        topology[s[0]][s[1]][s[2]] = True

    # calculate the relations, ie for a seq, the others seq that can be factorized with it
    relations = {}
    _computed = set()
    for seq in scripts:
        if not topology[seq[0]][seq[1]][seq[2]]:
            continue

        cubes = {e for e in _computed if
                 topology[e[0]][seq[1]][seq[2]] and
                 topology[seq[0]][e[1]][seq[2]] and
                 topology[seq[0]][seq[1]][e[2]]}

        for c in cubes:
            relations[c].add(seq)

        relations[seq] = cubes
        _computed.add(seq)

    def _neighbours(t1, t2):
        x1, y1, z1 = t1
        x2, y2, z2 = t2
        yield x1, y1, z1
        yield x1, y1, z2
        yield x1, y2, z1
        yield x1, y2, z2
        yield x2, y1, z1
        yield x2, y1, z2
        yield x2, y2, z1
        yield x2, y2, z2

    def _factors(candidate, factorisation):
        # sorting the list of candidate to get the one with the most of potential factors
        candidate.sort(key=lambda e: len(relations[e]), reverse=True)

        for r in candidate:
            _facto = set(it.chain.from_iterable(_neighbours(t, r) for t in factorisation))
            _candidate = set(candidate)
            for i in _facto:
                _candidate &= set(relations[i])

            if _candidate:
                yield from _factors(list(_candidate), _facto)
            else:
                yield _facto

        yield factorisation

    _candidate = [r for r in relations]
    _candidate.sort(key=lambda e: len(relations[e]))

    e = _candidate.pop()
    factorisations = next(iter(_factors(list(relations[e]), [e])))

    remaining = set(sequences) - set(scripts[f] for f in factorisations)
    factorisations = tuple(factor({primitives[i][seme] for seme in semes}) for i, semes in enumerate(zip(*factorisations)))

    if remaining:
        return [factorisations] + factor(remaining)
    else:
        return [factorisations]


def pack_factorisation(facto_list):
    """
    :param facto_list: list of parser or tuple of factorisation
    :return:
    """
    _sum = []
    for f in facto_list:
        if isinstance(f, Script):
            _sum.append(f)
        else:
            # tuple of factorisation
            _sum.append(MultiplicativeScript(children=(pack_factorisation(l_f) for l_f in f)))

    if len(_sum) == 1:
        return _sum[0]
    else:
        return AdditiveScript(children=_sum)


def factorize(script):
    if isinstance(script, Script):
        seqs = script.singular_sequences
    elif isinstance(script, list) or hasattr(script, '__iter__'):
        seqs = list(it.chain.from_iterable(s.singular_sequences for s in script))
    else:
        raise ValueError

    result = pack_factorisation(factor(seqs))
    return result


def inverse_relation(relation_name):
    from ..relations import INVERSE_RELATIONS
    return INVERSE_RELATIONS[relation_name]



from .terms import Term
from .dictionary import Dictionary
from .tools import term
from .table import Table
from .version import DictionaryVersion, latest_dictionary_version, create_dictionary_version, get_available_dictionary_version
from . import script


import logging
from collections import OrderedDict, defaultdict
from itertools import groupby, combinations, permutations, chain, repeat

import numpy as np
from scipy.sparse.coo import coo_matrix
from scipy.sparse.csr import csr_matrix
from scipy.sparse.dok import dok_matrix

from ieml.commons import cached_property
from ieml.dictionary.script.script import MultiplicativeScript, AdditiveScript, NullScript

logger = logging.getLogger(__name__)
RELATIONS = [
            'contains',         # 0
            'contained',        # 1
            'father_substance', # 2
            'child_substance',  # 3
            'father_attribute', # 4
            'child_attribute',  # 5
            'father_mode',      # 6
            'child_mode',       # 7
            'opposed',          # 8
            'associated',       # 9
            'crossed',          # 10
            'twin',             # 11
            'table_0',
            'table_1',
            'table_2',
            'table_3',
            'table_4',
            'table_5',

             # 'inclusion',        # 12
             # 'father',           # 13
             # 'child',            # 14
             # 'etymology',        # 15
             # 'siblings',         # 16
             # 'table'             # 17
             ]

INVERSE_RELATIONS = {
    'father_substance': 'child_substance',
    'child_substance': 'father_substance',  # 3
    'father_attribute': 'child_attribute', # 4
    'child_attribute': 'father_attribute',  # 5
    'father_mode': 'child_mode',      # 6
    'child_mode': 'father_mode',
    'contains': 'contained',
    'contained': 'contains',
    'opposed':'opposed',          # 8
    'associated':'associated',       # 9
    'crossed': 'crossed',        # 10
    'twin': 'twin',
    'table_0': 'table_0',
    'table_1': 'table_1',
    'table_2': 'table_2',
    'table_3': 'table_3',
    'table_4': 'table_4',
    'table_5': 'table_5',
    'father': 'child',
    'child': 'father',
    'inclusion': 'inclusion',
    'etymology': 'etymology',        # 15
    'siblings': 'siblings',         # 16
    'table': 'table'
}


class RelationsGraph:
    def __init__(self, dictionary):
        super().__init__()

        self.dictionary = dictionary
        self.relations = None
        self._compute_relations()

    def __getitem__(self, item):
        if isinstance(item, str) and item in RELATIONS:
            return self.relations[item]

        if isinstance(item, int):
            return self.relations[RELATIONS[item]]

        from .terms import Term
        if isinstance(item, Term):
            if item.dictionary != self.dictionary:
                raise ValueError("Invalid dictionary, (%s/%s)"%(str(item.dictionary), str(self.dictionary)))
            return Relations(term=item, relations_graph=self)

        raise NotImplemented

    def relation_type(self, term, relation_type):
        return [self.dictionary.index[j] for j in
                self.relations[relation_type][term.index, :].indices]

    def neighbours(self, term):
        return {
            reltype: self.relation_type(term, reltype) for reltype in RELATIONS
        }

    def _compute_relations(self):
        logger.log(logging.INFO, "Computing relations")

        self.relations = {}
        contains = self._compute_contains()
        self.relations['contains'] = csr_matrix(contains)
        self.relations['contained'] = csr_matrix(self.relations['contains'].transpose())

        father = self._compute_father()

        for i, r in enumerate(['_substance', '_attribute', '_mode']):
            self.relations['father' + r] = dok_matrix(father[i])

        siblings = self._compute_siblings()
        self.relations['opposed'] = dok_matrix(siblings[0])
        self.relations['associated'] = dok_matrix(siblings[1])
        self.relations['crossed'] = dok_matrix(siblings[2])
        self.relations['twin'] = dok_matrix(siblings[3])

        # self._do_inhibitions()

        for i, r in enumerate(['_substance', '_attribute', '_mode']):
            self.relations['child' + r] = self.relations['father' + r].transpose()

        # self.relations['siblings'] = sum(siblings)
        # self.relations['inclusion'] = np.clip(self.relations['contains'] + self.relations['contained'], 0, 1)
        # self.relations['father'] = self.relations['father_substance'] + \
        #                            self.relations['father_attribute'] + \
        #                            self.relations['father_mode']
        # self.relations['child'] = self.relations['child_substance'] + \
        #                           self.relations['child_attribute'] + \
        #                           self.relations['child_mode']
        # self.relations['etymology'] = self.relations['father'] + self.relations['child']

        table = self._compute_table_rank(self.relations['contained'])
        for i in range(6):
            self.relations['table_%d'%i] = table[i]

        missing = {s for s in RELATIONS if s not in self.relations}
        if missing:
            raise ValueError("Missing relations : {%s}"%", ".join(missing))

        self.relations = {reltype: csr_matrix(self.relations[reltype]) for reltype in RELATIONS}

    def _compute_table_rank(self, contained):
        logger.log(logging.DEBUG, "Computing tables relations")

        tables_rank = [([], []) for _ in range(6)]

        indices = [
            set(l) for l in np.split(contained.indices, contained.indptr)[1:-1]
        ]

        for root in self.dictionary.roots:
            for t0, t1 in combinations(self.dictionary.roots[root], 2):
                commons = [self.dictionary.index[i] for i in indices[t0.index] & indices[t1.index]]

                rank = max(map(lambda t: t.rank, commons))
                tables_rank[rank][0].extend((t0.index, t1.index))
                tables_rank[rank][1].extend((t1.index, t0.index))

        return [coo_matrix(([True]*len(i), (i, j)), shape=self.shape, dtype=np.bool) for i, j in tables_rank]

    def _compute_contains(self):
        logger.log(logging.DEBUG, "Computing contains/contained relations")
        # contain/contained

        # contains = coo_matrix(
        #     ([True]*len(self.dictionary), (, list(range(len(self.dictionary))))),
        #     shape=(len(self.dictionary), len(self.dictionary)), dtype=np.bool)
        #     # np.diag(np.ones(len(self.dictionary), dtype=np.bool))

        i = list(range(len(self.dictionary)))
        j = list(range(len(self.dictionary)))
        for r_p, v in self.dictionary.roots.items():
            paradigms = {t for t in v if t.script.paradigm}

            for p in paradigms:
                _contains = [self.dictionary.terms[ss].index for ss in p.script.singular_sequences] + \
                            [k.index for k in paradigms if k.script in p.script]
                i.extend(repeat(p.index, len(_contains)))
                j.extend(_contains)

        return coo_matrix(([True] * len(i), (i, j)), shape=self.shape, dtype=np.bool)

    def _compute_father(self):
        logger.log(logging.DEBUG, "Computing father/child relations")

        def _recurse_script(script):
            result = []
            for sub_s in script.children if isinstance(script, AdditiveScript) else [script]:
                if isinstance(sub_s, NullScript):
                    continue

                if sub_s in self.dictionary.terms:
                    result.append(self.dictionary.terms[sub_s].index)
                else:
                    if sub_s.layer > 0:
                        result.extend(chain.from_iterable(_recurse_script(c) for c in sub_s.children))

            return result

        # father = coo_matrix((3, len(self.dictionary), len(self.dictionary)), dtype=np.bool)

        father = [([], []) for _ in range(3)]

        for t in self.dictionary.terms.values():
            s = t.script

            for sub_s in s if isinstance(s, AdditiveScript) else [s]:
                if len(sub_s.children) == 0 or isinstance(sub_s, NullScript):
                    continue

                for i, s in enumerate(('father_substance', 'father_attribute', 'father_mode')):
                    if s in t.inhibitions:
                        continue

                    fathers_indexes = _recurse_script(sub_s.children[i])
                    father[i][0].extend(repeat(t.index, len(fathers_indexes)))
                    father[i][1].extend(fathers_indexes)

        return [coo_matrix(([True] * len(i), (i, j)), shape=self.shape, dtype=np.bool) for i, j in father]

    def _compute_siblings(self):
        # siblings
        # 1 dim => the sibling type
        #  -0 opposed
        #  -1 associated
        #  -2 crossed
        #  -3 twin
        def _opposed_sibling(s0, s1):
            return not s0.empty and not s1.empty and\
                   s0.cardinal == s1.cardinal and\
                   s0.children[0] == s1.children[1] and s0.children[1] == s1.children[0]

        def _associated_sibling(s0, s1):
            return s0.cardinal == s1.cardinal and\
                   s0.children[0] == s1.children[0] and \
                   s0.children[1] == s1.children[1] and \
                   s0.children[2] != s1.children[2]

        def _crossed_sibling(s0, s1):
            return s0.layer >= 2 and \
                   s0.cardinal == s1.cardinal and \
                   _opposed_sibling(s0.children[0], s1.children[0]) and \
                   _opposed_sibling(s0.children[1], s1.children[1])

        # siblings = coo_matrix((4, len(self.dictionary), len(self.dictionary)), dtype=np.bool)

        siblings = [([], []) for _ in range(4)]

        logger.log(logging.DEBUG, "Computing siblings relations")

        for root in self.dictionary.roots:
            _inhib_opposed = 'opposed' not in root.inhibitions
            _inhib_associated = 'associated' not in root.inhibitions
            _inhib_crossed = 'crossed' not in root.inhibitions
            _inhib_twin = 'twin' not in root.inhibitions

            if root.script.layer == 0:
                continue
            _twins = []

            for i, t0 in enumerate(self.dictionary.roots[root]):
                if not isinstance(t0.script, MultiplicativeScript):
                    continue

                if t0.script.children[0] == t0.script.children[1]:
                    _twins.append(t0)

                for t1 in [t for j, t in enumerate(self.dictionary.roots[root])
                           if j > i and isinstance(t.script, MultiplicativeScript)]:

                    if _inhib_opposed and _opposed_sibling(t0.script, t1.script):
                        siblings[0][0].extend((t0.index, t1.index))
                        siblings[0][1].extend((t1.index, t0.index))

                    if _inhib_associated and _associated_sibling(t0.script, t1.script):
                        siblings[1][0].extend((t0.index, t1.index))
                        siblings[1][1].extend((t1.index, t0.index))

                    if _inhib_crossed and _crossed_sibling(t0.script, t1.script):
                        siblings[2][0].extend((t0.index, t1.index))
                        siblings[2][1].extend((t1.index, t0.index))

            if _inhib_twin:
                _twins = sorted(_twins, key=lambda t: t.script.cardinal)
                for card, g in groupby(_twins, key=lambda t: t.script.cardinal):
                    twin_indexes = [t.index for t in g]

                    if len(twin_indexes) > 1:
                        index0, index1 = list(zip(*permutations(twin_indexes, r=2)))
                        siblings[3][0].extend(index0)
                        siblings[3][1].extend(index1)

                    # siblings[3, index0, index1] = True

        return [coo_matrix(([True]*len(i), (i, j)), shape=self.shape, dtype=np.bool) for i, j in siblings]

    @property
    def shape(self):
        return (len(self.dictionary), len(self.dictionary))

    def __setstate__(self, state):
        self.relations = state['relations']
        self.dictionary = state['dictionary']

    def __getstate__(self):
        return {
            'relations': self.relations,
            'dictionary': self.dictionary
        }


class Relations:
    def __init__(self, term, relations_graph):
        super().__init__()

        self.relations_graph = relations_graph
        self.term = term

    @cached_property
    def neighbours(self):
        rels = defaultdict(list)
        for reltype in RELATIONS:
            for t in self[reltype]:
                rels[t].append(reltype)

        neighbours = OrderedDict()

        for t in sorted(rels):
            neighbours[t] = rels[t]

        return neighbours

    def to(self, term, relations_types=None):
        if relations_types is None:
            relations_types = RELATIONS

        if term not in self.neighbours:
            return []

        return [reltype for reltype in self.neighbours[term] if reltype in relations_types]

    def __iter__(self):
        return self.neighbours.__iter__()

    def __len__(self):
        return len(self.neighbours)

    def __contains__(self, item):
        return item in self.neighbours

    def __getitem__(self, item):
        if isinstance(item, int):
            item = RELATIONS[item]

        if isinstance(item, str):
            return getattr(self, item)

        raise NotImplemented

    @property
    def father(self):
        return {
            's': self.father_substance,
            'a': self.father_attribute,
            'm': self.father_mode,
        }

    @property
    def child(self):
        return {
            's': self.child_substance,
            'a': self.child_attribute,
            'm': self.child_mode,
        }


def get_relation(reltype):
    def getter(self):
        return tuple(self.relations_graph.relation_type(term=self.term, relation_type=reltype))

    getter.__name__ = reltype
    return getter


for reltype in {'contains',
                'contained',
                'father_substance',
                'child_substance',
                'father_attribute',
                'child_attribute',
                'father_mode',
                'child_mode',
                'opposed',
                'associated',
                'crossed',
                'twin',
                'table_0',
                'table_1',
                'table_2',
                'table_3',
                'table_4',
                'table_5'}:
    setattr(Relations, reltype, cached_property(get_relation(reltype)))



import random

from ieml.syntax.commons import IEMLSyntax, IEMLSyntaxType
from ..syntax import Sentence, SuperSentence, Text, Word, Morpheme
from ieml.tools import ieml
from ..tools import RandomPoolIEMLObjectGenerator
from ..dictionary import Term

from .paths import path, resolve_ieml_object
from .parser import USLParser
from .usl import Usl


def usl(arg):
    if isinstance(arg, Usl):
        return arg
    if isinstance(arg, IEMLSyntax):
        if isinstance(arg, Term):
            return Usl(Word(root=Morpheme([arg])))
        return Usl(arg)
    if isinstance(arg, str):
        return USLParser().parse(arg)

    if isinstance(arg, dict):
        # map path -> Ieml_object
        return Usl(resolve_ieml_object(arg))

    try:
        rules = [(a, b) for a, b in arg]
    except TypeError:
        pass
    else:
        rules = [(path(a), ieml(b)) for a, b in rules]
        return Usl(resolve_ieml_object(rules))

    raise ValueError("Invalid argument to create an usl object.")

_ieml_objects_types = [Term, Word, Sentence, SuperSentence]
_ieml_object_generator = None

def random_usl(rank_type=None):
    global _ieml_object_generator
    if _ieml_object_generator is None:
        _ieml_object_generator = RandomPoolIEMLObjectGenerator(level=Text)

    if rank_type and not isinstance(rank_type, IEMLSyntaxType):
        raise ValueError('The wanted type for the generated usl object must be a IEMLType, here : '
                         '%s'%rank_type.__class__.__name__)

    if not rank_type:
        i = random.randint(0, 10)
        if i < 4:
            rank_type = _ieml_objects_types[i]
        else:
            rank_type = Text

    return usl(_ieml_object_generator.from_type(rank_type))


def replace_paths(u, rules):
    k = [(p,t) for p, t in {
            **usl(u).paths,
            **{path(p): ieml(t) for p, t in rules.items()}}.items()]
    return usl(k)



from .words import Morpheme, Word
from .sentences import Clause, Sentence, SuperClause, SuperSentence
from .texts import Text
from .hypertexts import Hyperlink, Hypertext, PropositionPath
from .parser import IEMLParser


import ply.lex as lxr
import logging

from ..constants import COORDINATES_KINDS

logger = logging.getLogger(__name__)

tokens = (
   'COORD_KIND',
   'COORD_INDEX',
   'PLUS',
   'LPAREN',
   'RPAREN',
   'COLON'
)


def get_lexer(module=None):
    t_COORD_KIND = r'[%s]'%''.join(COORDINATES_KINDS)
    t_COORD_INDEX = r'\d+'
    t_PLUS   = r'\+'
    t_LPAREN  = r'\('
    t_RPAREN  = r'\)'
    t_COLON = r':'
    t_ignore  = ' \t\n'

    # Error handling rule
    def t_error(t):
        logger.log(logging.ERROR, "Illegal character '%s'" % t.value[0])
        t.lexer.skip(1)

    return lxr.lex(module=module, errorlog=logging)



from itertools import chain

from .commons import IEMLSyntax
from ..exceptions import InvalidIEMLObjectArgument
from .sentences import Sentence, SuperSentence
from .words import Word


class Text(IEMLSyntax):
    closable = True

    def __init__(self, children):
        try:
            _children = [e for e in children]
        except TypeError:
            raise InvalidIEMLObjectArgument(Text, "The argument %s is not iterable." % str(children))

        if not all(isinstance(e, (Word, Sentence, SuperSentence, Text)) for e in _children):
            raise InvalidIEMLObjectArgument(Text, "Invalid type instance in the list of a text,"
                                                  " must be Word, Sentence, SuperSentence or Text")

        _children = list(chain([c for c in _children if not isinstance(c, Text)],
                         *(c.children for c in _children if isinstance(c, Text))))

        super().__init__(sorted(set(_children)))

    def compute_str(self, children_str):
        return '{/' + '//'.join(children_str) + '/}'

    def __contains__(self, item):
        if isinstance(item, Text):
            return all(c in self.children for c in item.children)
        elif isinstance(item,  (Word, Sentence, SuperSentence)):
            return item in self.children

        raise NotImplemented


from collections import defaultdict
from itertools import chain

from ieml.dictionary.table import Cell, Table3D, TableSet
from ieml.dictionary.terms import Term

# from ..tools import ieml
# from .template import Template

# template = Template(ieml("[([M:M:.a.-]+[M:.-',M:.-',S:.-'B:.-'n.-S:.U:.-',_])*([E:U:T:.]+[E:U:.wa.-])]"), ['r0', 'r1'])
# collection = list(template)
# inverse_terms = defaultdict(list)
#
# for u in collection:
#     for t in u.paths.values():
#         inverse_terms[t].append(u)
#
# inverse_root = defaultdict(list)
# for t in inverse_terms:
#     inverse_root[t.root].extend(inverse_terms[t])
#
# roots = sorted(inverse_root, key=lambda t: len(inverse_root[t]), reverse=True)


def project_usls_on_dictionary(usls, allowed_terms=None):
    """`usls` is an iterable of usl.

    return a mapping term -> usl list
    """

    cells_to_usls = defaultdict(set)
    tables = set()

    for u in usls:
        for t in u.objects(Term):
            for c in t.singular_sequences:
                # This is the first time we meet the cell c
                if not cells_to_usls[c]:
                    tables.update(c.relations.contained)

                cells_to_usls[c].add(u)

    if allowed_terms:
        allowed_terms = set(allowed_terms)
        tables = tables & allowed_terms
        cells_to_usls = {c: l for c, l in cells_to_usls.items() if c in allowed_terms}

    tables_to_usls = {
        table: list(set(u for c in table.singular_sequences for u in cells_to_usls[c]))
            for table in tables if not isinstance(table, TableSet)
    }

    return tables_to_usls


def project_usl_with_data(usls_data, metric=None):
    """
    usls_data: usl => data[]
    :param usls_data:
    :return:
    """

    projection = project_usls_on_dictionary(usls_data)
    all_terms = set(c for u in usls_data for t in u.objects(Term) for c in t.singular_sequences)
    if metric is None:
        metric = lambda e: len(e['posts']) * len(all_terms.intersection(e['table'].singular_sequences))

    return sorted(({
        'table': table,
        'usls': usls,
        'posts': list(set(chain.from_iterable(usls_data[u] for u in usls)))
    } for table, usls in projection.items()), key=metric, reverse=True)


import copy
import logging
import pickle
from functools import lru_cache
from urllib.request import urlretrieve
import datetime
import urllib.parse
import json
import os
import re

from ieml.dictionary.relations import RelationsGraph
from .. import get_configuration, ieml_folder
from ..constants import LANGUAGES

logger = logging.getLogger(__name__)
VERSIONS_FOLDER = os.path.join(ieml_folder, get_configuration().get('VERSIONS', 'versionsfolder'))

if not os.path.isdir(VERSIONS_FOLDER):
    os.mkdir(VERSIONS_FOLDER)


def get_available_dictionary_version():
    version_url = get_configuration().get('VERSIONS', 'versionsurl')
    from ieml.tools import list_bucket
    return [v[:-5] for v in list_bucket(version_url)]


def latest_dictionary_version():
    return DictionaryVersion(get_available_dictionary_version()[0])


def _date_to_str(date):
    return date.strftime('%Y-%m-%d_%H:%M:%S')


def _str_to_date(string):
    return datetime.datetime.strptime(string, '%Y-%m-%d_%H:%M:%S')


class DictionaryVersionSingleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        date = args[0] if len(args) == 1 else kwargs['date']

        if date is None:
            date = get_configuration().get('VERSIONS', 'defaultversion')

        if isinstance(date, DictionaryVersion):
            return date

        if isinstance(date, str):
            if date.startswith('dictionary_'):
                date = _str_to_date(date.split('.')[0].split('_', maxsplit=1)[1])
            else:
                date = _str_to_date(date)
        elif isinstance(date, datetime.date):
            date = _str_to_date(_date_to_str(date))
        else:
            raise ValueError("Invalid date format for dictionary version %s." % _date_to_str(date))

        if date not in cls._instances:
            cls._instances[date] = super(DictionaryVersionSingleton, cls).__call__(date)

        return cls._instances[date]


class DictionaryVersion(metaclass=DictionaryVersionSingleton):
    """
    Track the available versions
    """
    def __init__(self, date):
        super(DictionaryVersion, self).__init__()

        self.date = date

        # A list of str. All the script defined in this dictionary.
        self.terms = None

        # The list of root paradigms defined
        self.roots = None

        # A map root -> list of str. The list of relations to inhibate in this root paradigm
        self.inhibitions = None

        # A map term str -> map fr|en -> str
        # The translations of each terms in fr and en (can add more languages)
        self.translations = None

        # A map version str -> map old term str -> new term str
        # To translate old dictionary terms into new ones.
        # Each keys is a version that modify terms (change theirs script)
        # The change must be applied in chronological order (from the old version to new one)
        self.diff = None

        self.loaded = False

    def __str__(self):
        return 'dictionary_%s' % _date_to_str(self.date)

    def __getstate__(self):
        self.load()

        return {
            'version': _date_to_str(self.date),
            'terms': self.terms,
            'roots': self.roots,
            'inhibitions': self.inhibitions,
            'translations': self.translations,
            'diff': self.diff
        }

    def __setstate__(self, state):
        self.date = _str_to_date(state['version'])
        self.terms = state['terms']
        self.roots = state['roots']
        self.inhibitions = state['inhibitions']
        self.translations = state['translations']
        self.diff = state['diff'] if 'diff' in state else {}

        self.loaded = True

    def json(self):
        return json.dumps(self.__getstate__())

    def load(self):
        """
        Download the dictionary version and cache the retrieved file.
        :return: None
        """
        if self.loaded:
            return

        file_name = "%s.json" % str(self)
        file = os.path.join(VERSIONS_FOLDER, file_name)

        if not os.path.isfile(file):
            DICTIONARY_BUCKET_URL = get_configuration().get('VERSIONS', 'versionsurl')
            url = urllib.parse.urljoin(DICTIONARY_BUCKET_URL, file_name)
            logger.log(logging.INFO, "Downloading dictionary %s at %s" % (file_name, url))

            urlretrieve(url, file)

        with open(file, 'r') as fp:
            self.__setstate__(json.load(fp))

    def __eq__(self, other):
        return self.date == other.date

    def __hash__(self):
        return str(self).__hash__()

    def __lt__(self, other):
        return self.date.__lt__(other.date)

    def __gt__(self, other):
        return self.date.__gt__(other.date)

    def __le__(self, other):
        return self.date.__le__(other.date)

    def __ge__(self, other):
        return self.date.__ge__(other.date)

    @property
    def cache(self):
        file_name = "cache_%s.pkl" % str(self)
        return os.path.join(VERSIONS_FOLDER, file_name)

    @property
    def is_cached(self):
        return os.path.isfile(self.cache)

    @lru_cache(5)
    def diff_for_version(self, older_version):
        older_version.load()
        self.load()

        result = {
            sc: sc for sc in older_version.terms
        }

        chronology = sorted(filter(lambda v: v >= older_version, (DictionaryVersion(v) for v in self.diff)))

        for v in chronology:
            for sc_old, sc_new in result.items():
                if sc_new in self.diff[str(v)]:
                    result[sc_old] = self.diff[str(v)][sc_new]

        return result


def _latest_installed_version():
    version_file_pattern = re.compile("^dictionary_\d{4}-\d{2}-\d{2}_\d{2}:\d{2}:\d{2}\.json")

    all_versions = sorted((file for file in os.listdir(VERSIONS_FOLDER) if version_file_pattern.match(file)))

    if all_versions:
        return DictionaryVersion(all_versions[-1])
    else:
        return None


_default_version = _latest_installed_version()


def get_default_dictionary_version():
    """
    Return the default dictionary version. The version is the latest version installed on
    the system. If there is none installed yet, it download and install the latest version
    available online.
    To upgrade to a newer version, you can run `set_default_dictionary_version(latest_dictionary_version())`

    :return: the default dictionary version
    """
    if _default_version is None:
        set_default_dictionary_version(latest_dictionary_version())

    return _default_version


def set_default_dictionary_version(version):
    global _default_version
    if not isinstance(version, DictionaryVersion):
        version = DictionaryVersion(version)

    _default_version = version


def create_dictionary_version(old_version=None, add=None, update=None, remove=None):
    """

    :param old_version: the dictionary version to build the new version from
    :param add: a dict with the element to add {'terms': list of script to add,
                                                'roots': list of script to add root paradigm,
                                                'inhibitions': dict {root_p: list of relations to inhibits in this root p}
                                                'translations': dict {language: {script: traduction}}}
    :param update: a dict to update the translations and inhibtions or the terms (new mapping)
            map terms|inhibitions|translations -> old -> new
    :param remove: a list of term to remove, they are removed from root, terms, inhibitions and translations
    :return:
    """
    v = latest_dictionary_version()
    last_date = v.date

    while True:
        new_date = datetime.datetime.utcnow()
        if new_date != last_date:
            break

    if old_version is None:
        old_version = v

    old_version.load()

    state = {
        'version': _date_to_str(new_date),
        'terms': copy.deepcopy(old_version.terms),
        'roots': copy.deepcopy(old_version.roots),
        'inhibitions': copy.deepcopy(old_version.inhibitions),
        'translations': copy.deepcopy(old_version.translations),
        'diff': {**copy.deepcopy(old_version.diff),
                 str(old_version): {}}
    }

    # if merge is not None:
    #     for m_version in merge:
    #         m_version.load()
    #
    #         terms_to_add = set(m_version.terms).difference(state['terms'])
    #         roots_to_add = set(m_version.roots).difference(state['roots'])
    #
    #         state['terms'].extend(terms_to_add)
    #         state['roots'].extend(roots_to_add)
    #         state['inhibitions'].update({r: m_version.inhibitions[r] for r in roots_to_add if r in m_version.inhibitions})
    #         for l in LANGUAGES:
    #             state['translations'][l].update({s: m_version.translations[l][s] for s in terms_to_add})

    if remove is not None:
        state['terms'] = list(set(state['terms']).difference(remove))
        state['roots'] = list(set(state['roots']).difference(remove))
        for r in remove:
            if r in state['inhibitions']:
                del state['inhibitions'][r]

            for l in LANGUAGES:
                if r in state['translations'][l]:
                    del state['translations'][l][r]

            state['diff'][str(old_version)][r] = None

    if add is not None:
        if 'terms' in add:
            state['terms'] = list(set(state['terms']).union(add['terms']))
        if 'roots' in add:
            state['roots'] = list(set(state['roots']).union(add['roots']))
        if 'inhibitions' in add:
            if set(state['inhibitions']).intersection(set(add['inhibitions'])):
                raise ValueError("Error in creating a new dictionary versions, trying to add multiples "
                                 "inhibitions rules for the same script.")

            state['inhibitions'] = {**state['inhibitions'], **add['inhibitions']}
        if 'translations' in add:
            if any(set(state['translations'][l]).intersection(set(add['translations'][l])) for l in LANGUAGES):
                raise ValueError("Error in creating a new dictionary version, trying to add multiples "
                                 "translation for the script {%s}. Those script may already exists in the dictionary."%', '.join(['"%s": [%s]'%(l, ', '.join('"%s"'%str(t) for t in set(state['translations'][l]).intersection(set(add['translations'][l])))) for l in LANGUAGES]))

            state['translations'] = {l: {**state['translations'][l], **add['translations'][l]} for l in LANGUAGES}

    if update is not None:
        if 'inhibitions' in update:
            for s, l in update['inhibitions'].items():
                if s not in state['inhibitions']:
                    continue
                state['inhibitions'][s] = l

        if 'translations' in update:
            state['translations'] = {l: {**state['translations'][l], **update['translations'][l]} for l in LANGUAGES}

        if 'terms' in update:
            state['terms'] = set(t for t in state['terms'] if t not in update['terms'])

            roots = set(state['roots']).intersection(update['terms'])
            state['roots'] = set(t for t in state['roots'] if t not in update['terms'])

            for t_old in update['terms']:
                t_new = update['terms'][t_old]

                state['diff'][str(old_version)][t_old] = t_new
                state['terms'].add(t_new)

                if t_old in roots:
                    state['roots'].add(t_new)

                for l in LANGUAGES:
                    state['translations'][l][t_new] = state['translations'][l][t_old]
                    del state['translations'][l][t_old]

                if t_old in state['inhibitions']:
                    state['inhibitions'][t_new] = state['inhibitions'][t_old]
                    del state['inhibitions'][t_old]

    dictionary_version = DictionaryVersion(new_date)
    dictionary_version.__setstate__(state)

    from ieml.dictionary import Dictionary

    if set(old_version.terms) == set(state['terms']) and set(old_version.roots) == set(state['roots']) and \
       all(old_version.inhibitions[s] == state['inhibitions'][s] for s in old_version.inhibitions):

        old_dict_state = Dictionary(old_version).__getstate__()

        d = Dictionary.__new__(Dictionary)
        rel_graph = RelationsGraph.__new__(RelationsGraph)
        rel_graph.__setstate__({
            'dictionary': d,
            'relations': old_dict_state['relations'].__getstate__()['relations']
        })

        state = {
            'version': dictionary_version,
            'relations': rel_graph,
            'scripts': old_dict_state['scripts'],
        }

        d.__setstate__(state)
        save_dictionary_to_cache(d)
    else:
        # graph is updated, must check the coherence
        Dictionary(dictionary_version)

    return dictionary_version


def save_dictionary_to_cache(dictionary):
    logger.log(logging.INFO, "Saving dictionary cache to disk (%s)" % dictionary.version.cache)

    with open(dictionary.version.cache, 'wb') as fp:
        pickle.dump(dictionary, fp, protocol=4)


def load_dictionary_from_cache(version):
    logger.log(logging.INFO, "Loading dictionary from disk (%s)" % version.cache)

    with open(version.cache, 'rb') as fp:
        return pickle.load(fp)



import logging
import os
import ply.yacc as yacc
import threading

from ieml import parser_folder
from ieml.exceptions import CannotParse
from ieml.syntax.parser import IEMLParser
from ieml.usl.parser.lexer import tokens, get_lexer
from ieml.usl import Usl
from ieml.commons import Singleton


class USLParser(metaclass=Singleton):
    tokens = tokens
    lock = threading.Lock()

    def __init__(self):

        # Build the lexer and parser
        self.lexer = get_lexer()
        self.parser = yacc.yacc(module=self, errorlog=logging, start='usl',
                                debug=False, optimize=True, picklefile=os.path.join(parser_folder, "usl_parser.pickle"))

    def parse(self, s):
        """Parses the input string, and returns a reference to the created AST's root"""
        # self.usl = s
        # self.root = None
        with self.lock:
            try:
                return self.parser.parse(s, lexer=self.lexer)
            except CannotParse as e:
                e.s = s
                raise e

        #
        # if self.root is not None:
        #     return self.root
        # else:
        #     raise CannotParse(s, "Invalid usl.")

    # Parsing rules
    def p_usl(self, p):
        """ usl : IEML_OBJECT """
        p[0] = Usl(IEMLParser().parse(p[1]))

    def p_error(self, p):
        if p:
            msg = "Syntax error at '%s' (%d, %d)" % (p.value, p.lineno, p.lexpos)
        else:
            msg = "Syntax error at EOF"

        raise CannotParse(None, msg)




import logging
from functools import lru_cache
import threading

from ply import yacc

from ....exceptions import CannotParse
from .lexer import tokens, get_lexer
from ..paths import Coordinate, AdditivePath, MultiplicativePath, ContextPath
from ....commons import Singleton


class PathParser(metaclass=Singleton):
    tokens = tokens
    lock = threading.Lock()

    def __init__(self):

        # Build the lexer and parser
        self.lexer = get_lexer()
        self.parser = yacc.yacc(module=self, errorlog=logging, start='path',
                                debug=False, optimize=True, picklefile="parser/path_parser.pickle")
        # rename the parsing method (can't name it directly parse with lru_cache due to ply checking)
        self.parse = self.t_parse

    @lru_cache(maxsize=1024)
    def t_parse(self, s):
        """Parses the input string, and returns a reference to the created AST's root"""
        # self.root = None
        # self.path = s
        with self.lock:
            try:
                return self.parser.parse(s, lexer=self.lexer, debug=False)
            except CannotParse as e:
                e.s = s
                raise e

        # if self.root is not None:
        #     if len(self.root.children) == 1:
        #         self.root = self.root.children[0]
        #
        #     return self.root
        # else:
        #     raise CannotParse(s, "Invalid path.")

    def p_path(self, p):
        """path : additive_path"""
        if len(p[1].children) == 1:
            p[0] = p[1].children[0]
        else:
            p[0] = p[1]

    def p_additive_path(self, p):
        """ additive_path : path_sum"""
        p[0] = AdditivePath(p[1])

    def p_path_sum(self, p):
        """ path_sum : ctx_path
                  | path_sum PLUS ctx_path"""
        if len(p) == 2:
            p[0] = [p[1]]
        else:
            p[0] = p[1] + [p[3]]

    def p_ctx_path(self, p):
        """ ctx_path : ctx_coords"""
        if len(p[1]) == 1:
            p[0] = p[1][0]
        else:
            p[0] = ContextPath(p[1])

    def p_ctx_coords(self, p):
        """ ctx_coords : multiplicative_path
                        | ctx_coords COLON multiplicative_path"""
        if len(p) == 2:
            p[0] = [p[1]]
        else:
            p[0] = p[1] + [p[3]]

    def p_multiplicative_path(self, p):
        """ multiplicative_path : product"""
        p[0] = MultiplicativePath(p[1])

    def p_product(self, p):
        """ product : additive_path_p
                    | coordinate
                    | product additive_path_p
                    | product coordinate"""
        if len(p) == 2:
            p[0] = [p[1]]
        else:
            p[0] = p[1] + [p[2]]

    def p_additive_path_p(self, p):
        """ additive_path_p : LPAREN additive_path RPAREN"""
        p[0] = p[2]

    def p_coordinate(self, p):
        """ coordinate : COORD_KIND
                        | COORD_KIND COORD_INDEX"""

        if len(p) == 2:
            p[0] = Coordinate(p[1])
        else:
            p[0] = Coordinate(p[1], int(p[2]))

    def p_error(self, p):
        if p:
            msg = "Syntax error at '%s' (%d, %d)" % (p.value, p.lineno, p.lexpos)
        else:
            msg = "Syntax error at EOF"

        raise CannotParse(None, msg)




class cached_property:
    def __init__(self, factory):
        self._factory = factory
        self._attr_name = factory.__name__

    def __get__(self, instance, owner):
        attr = self._factory(instance)
        setattr(instance, self._attr_name, attr)
        return attr


class TreeStructure:
    def __init__(self):
        super().__init__()
        self._str = None
        self._paths = None
        self.children = None  # will be an iterable (list or tuple)

    def __str__(self):
        return self._str

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        if not isinstance(other, (TreeStructure, str)):
            return False

        return self._str == str(other)

    def __hash__(self):
        """Since the IEML string for any proposition AST is supposed to be unique, it can be used as a hash"""
        return self.__str__().__hash__()

    def __iter__(self):
        """Enables the syntactic sugar of iterating directly on an element without accessing "children" """
        return self.children.__iter__()

    def tree_iter(self):
        yield self
        for c in self.children:
            yield from c.tree_iter()


def fullname(cls):
    return cls.__module__ + '.' + cls.__qualname__


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            # this code is to clean up duplicate class if we reload modules
            to_remove = [i for i in cls._instances if fullname(i) == fullname(cls)]
            for i in to_remove:
                del cls._instances[i]

            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)

        return cls._instances[cls]


from functools import partial
from itertools import chain, groupby


from ieml.dictionary.distance import get_matrix
from ieml.dictionary.version import latest_dictionary_version


def w_associate(w0, w1):
    return w0.root == w1.root


def w_opposed(w0, w1):
    return w0.root == w1.flexing and w0.flexing == w1.root


def m_opposed(m0, m1):
    return len(m0) == len(m1) == 1 and m1[0] in m0[0].relations.opposed


def w_crossed(w0, w1):
    return m_opposed(w0.root, w1.root) and m_opposed(w0.flexing, w1.flexing)


def count_id(w0):
    """
    0 -> no terms idd
    1 -> most term idd are shared in root morphem
    2 -> most term idd are shared in flexing morphem
    3 -> most term idd are shared root <-> flexing (crossed)
    :param w0:
    :param w1:
    :return:
    """

    def f(w1):
        count = [set(w0.root).intersection(w1.root),
                 set(w0.flexing).intersection(w1.flexing),
                 set(w0.root).intersection(w1.flexing) | set(w1.root).intersection(w0.flexing)]

        if any(count):
            return max((1,2,3), key=lambda i: len(count[i - 1]))
        else:
            return 0

    return f

relations = get_matrix('relation', latest_dictionary_version())


def count_relations(w0):
    """
    0 -> no terms idd
    1 -> most term idd are shared in root morphem
    2 -> most term idd are shared in flexing morphem
    3 -> most term idd are shared root <-> flexing (crossed)
    :param w0:
    :param w1:
    :return:
    """

    root_w0_relations = set(chain.from_iterable(relations[t.index, :].indices for t in w0.root))
    flexing_w0_relations = set(chain.from_iterable(relations[t.index, :].indices for t in w0.flexing))

    def f(w1):
        root_w1 = set(t.index for t in w1.root)
        flexing_w1 = set(t.index for t in w1.flexing)

        count = [root_w0_relations.intersection(root_w1),
                 flexing_w0_relations.intersection(flexing_w1),
                 root_w0_relations.intersection(flexing_w1) | flexing_w0_relations.intersection(root_w1)]

        if any(count):
            return max((1,2,3), key=lambda i: len(count[i - 1]))
        else:
            return 0
    return f

def w_twin(w0):
    return w0.root == w0.flexing


def order_word_list(w0, word_list):
    klass = {'A': [partial(w_associate, w0=w0), partial(w_opposed, w0=w0), partial(w_crossed, w0=w0)],
             'B': count_id(w0),
             'C': count_relations(w0),
             'D': w_twin}

    SUB_KLASS = {1: '0', 2: '1', 3: '2'}

    def get_mark(w):
        for k in ('A', 'B', 'C', 'D'):
            if isinstance(klass[k], list):
                for i, f in enumerate(klass[k]):
                    if f(w1=w):
                        return k + SUB_KLASS[i + 1]
            else:
                o = klass[k](w)
                if o != 0:
                    return k + SUB_KLASS[o]

        return 'E'

    res = sorted(((w, get_mark(w)) for w in word_list), key=lambda k: k[1])
    return {k: [w[0] for w in v] for k, v in groupby(res,  key=lambda k: k[1])}


def display_ranking(w0, word_list):
    print("Ranking from %s"%str(w0))

    res = order_word_list(w0, word_list)
    for k in sorted(res):
        print("-- Rank %s --"%k)
        for w in res[k]:
            print('\t %s'%str(w))





import logging
import os
import pickle
from enum import Enum, unique, IntEnum
from itertools import combinations, product, groupby, chain

import bidict
import numpy as np
from scipy.sparse.csr import csr_matrix

from ieml.dictionary.dictionary import Dictionary
from ieml.dictionary.tools import term

logger = logging.getLogger(__name__)

RelationType = unique(IntEnum('RelationType', {
    'Null': 0,
    'Equal': 1,
    'Crossed': 2,
    'Associated': 3,
    'Twin': 4,
    'Opposed': 5,
    **{'Rank_%d'%i: 6 + i  for i in range(6)},

    **{'Child_%s'%''.join(s): int(11 + (3 ** i - 1) / 2 + j) for i in range(1, 4) for j, s in
       enumerate(product('sam', repeat=i))},

    **{'Father_%s' % ''.join(s): int(50 + (3 ** i - 1) / 2 + j) for i in range(1, 4) for j, s in
       enumerate(product('sam', repeat=i))}
}))


def default_metric(dictionary_version):
    mat = get_matrix('distance', dictionary_version)
    return lambda t0, t1: mat[t0.index, t1.index]


def term_ranking(t):
    order = get_matrix('order', t.dictionary.version)
    terms = [term(int(i)) for i in sorted(order[t.index, :].indices, key=lambda o: order[t.index, o])]
    return [t for t in terms if len(t) == 1], [t for t in terms if len(t) != 1]


def relations_pack(t0, term_list):
    reltypes = get_matrix('relation', t0.dictionary.version)

    res = [(reltypes[t0.index, t1.index], t1) for t1 in term_list]

    return [(RelationType(key).name, [t[1] for t in v]) for key, v in groupby(res, key=lambda t: t[0])]


def test_metric(t0):
    def _str_term(t):
        return "%s - (%s)"%(str(t), t.translations['fr'])

    print("Ranking from %s"%_str_term(t0))
    ss, para = term_ranking(t0)
    ss_pack = relations_pack(t0, ss)
    para_pack = relations_pack(t0, para)

    def _print_pack(p):
        for reltype, v in p:
            print("- %s -"%str(reltype))
            print("\n".join(_str_term(t) for t in v))

    print("SS:")
    _print_pack(ss_pack)
    print("\nParadigms")
    _print_pack(para_pack)
    # print('\n'.join("[%.3f] [%s]: %s"%(r[0], RelationType(reltypes[t0.index, r[1].index]).name,_str_term(r[1])) for r in res))




def get_matrix(name, version):
    file = '/tmp/cache_%s_%s.npy' % (name, str(version))
    if os.path.isfile(file):
        with open(file, 'rb') as fp:
            return pickle.load(fp)
    else:
        logger.log(logging.INFO, "Building distance matrix '%s'."%name)
        mat = MATRIX_BUILD[name](version)
        for k, v in mat.items():
            file_name = '/tmp/cache_%s_%s.npy' % (k, str(version))
            with open(file_name, 'wb') as fp:
                pickle.dump(v, fp)

        return mat[name]


def get_relation(t0, t1, prefix=None):
    if prefix is None:
        raise NotImplemented
        # reltype = min({r for r in t0.relations.to(t1) if r not in ('contained', 'contains')},
        #                   key=lambda r: RELATIONS_TYPES[r])
        #
        # if reltype == 'associated':
        #     return RelationType.Associated
        # elif reltype == 'opposed':
        #     return RelationType.Opposed
        # elif reltype == 'crossed':
        #     return RelationType.Crossed
        # elif reltype == 'twin':
        #     return RelationType.Twin
        # elif reltype.startswith('table_'):
        #     return RelationType['Rank_%d'%int(reltype[6:7])]
        # else:
        #     raise NotImplemented
    else:
        type = 'Child' if t0.layer < t1.layer else 'Father'
        return RelationType['%s_%s'%(type, prefix)]


RELATION_ORDER_FROM_MAX_RANK = {
    i: ['Equal', 'Associated', 'Opposed', 'Crossed', 'Twin', 'Father1'] for i in range(6)
}

RELATION_ORDER_FROM_MAX_RANK[0] += ['Rank_0', 'Child1', 'Father2', 'Child2', 'Father3', 'Child3', 'Null']
RELATION_ORDER_FROM_MAX_RANK[1] += ['Rank_1', 'Child1', 'Father2', 'Child2', 'Rank_0','Father3', 'Child3', 'Null']
RELATION_ORDER_FROM_MAX_RANK[2] += ['Rank_2', 'Child1', 'Father2', 'Rank_1', 'Child2', 'Rank_0', 'Father3', 'Child3', 'Null']
RELATION_ORDER_FROM_MAX_RANK[3] += ['Rank_3', 'Child1', 'Rank_2', 'Father2', 'Rank_1', 'Child2', 'Rank_0', 'Father3', 'Child3', 'Null']
RELATION_ORDER_FROM_MAX_RANK[4] += ['Rank_4', 'Child1', 'Rank_3', 'Rank_2', 'Father2', 'Rank_1', 'Child2', 'Rank_0', 'Father3', 'Child3', 'Null']
RELATION_ORDER_FROM_MAX_RANK[5] += ['Rank_5', 'Child1', 'Rank_4', 'Rank_3', 'Rank_2', 'Father2', 'Rank_1', 'Child2', 'Rank_0', 'Father3', 'Child3', 'Null']

for i, rellist in RELATION_ORDER_FROM_MAX_RANK.items():
    res = []
    for rel in rellist:
        if rel.startswith('Child'):
            res += [RelationType['Child_%s'%''.join(k)] for k in product('sam', repeat=int(rel[5:6]))]
        elif rel.startswith('Father'):
            res += [RelationType['Father_%s'%''.join(k)] for k in product('sam', repeat=int(rel[6:7]))]
        else:
            res += [RelationType[rel]]

    RELATION_ORDER_FROM_MAX_RANK[i] = {rel: j for j, rel in enumerate(res)}


def get_relation_value(relation, t0):
    return RELATION_ORDER_FROM_MAX_RANK[t0.max_rank][relation]


RELATIONS_TYPES = [
    ('associated', RelationType.Associated),
    ('opposed', RelationType.Opposed),
    ('crossed', RelationType.Crossed),
    ('twin', RelationType.Twin),
    ('table_5', RelationType.Rank_5),
    ('table_4', RelationType.Rank_4),
    ('table_3', RelationType.Rank_3),
    ('table_2', RelationType.Rank_2),
    ('table_1', RelationType.Rank_1),
    ('table_0', RelationType.Rank_0)
]

def _build_distance_matrix(version):
    def _enumerate_ancestors(t, prefix='', seen=None):
        if seen is None:
            seen = set()
        for k, v in t.relations.father.items():
            for t1 in v:
                # if t1 is layer 0, we include this etymology only if it is a direct father/child
                if t1.layer == 0 and len(prefix) != 0:
                    continue

                if t1 not in seen:
                    yield (prefix + k, t1)
                    seen.add(t1)

                if len(prefix) < 2:
                    yield from _enumerate_ancestors(t1, prefix=prefix + k, seen=seen)

    d = Dictionary(version)

    def _put(mat,d, i, j):
        mat[0].extend(d)
        mat[1].extend(i)
        mat[2].extend(j)

    order_matrix = ([], [], [])
    relation_type_matrix = ([], [], [])

    all_indices = {
        rel : [set(l) for l in np.split(d.relations_graph[rel].indices, d.relations_graph[rel].indptr)[1:-1]]
        for rel, _ in RELATIONS_TYPES
    }

    for root in d.roots:
        past = set()
        for t0 in root.relations.contains:
            past.add(t0.index)
            seen = set(past)

            for rel_graph, rel_type in RELATIONS_TYPES:
                indices = all_indices[rel_graph][t0.index].difference(seen)
                if indices:
                    value = get_relation_value(rel_type, t0)

                    _put(order_matrix, [value] * len(indices), [t0.index] * len(indices), indices)
                    _put(order_matrix, [value] * len(indices), indices, [t0.index] * len(indices))

                    _put(relation_type_matrix, [int(rel_type)] * len(indices), [t0.index] * len(indices), indices)
                    _put(relation_type_matrix, [int(rel_type)] * len(indices), indices, [t0.index] * len(indices))

                    seen.update(indices)

    for layer in d.layers:
        for t0 in layer:
            for prefix, t1 in _enumerate_ancestors(t0):
                rel = get_relation(t0, t1, prefix=prefix)
                order = get_relation_value(rel, t0)

                _put(relation_type_matrix, [int(rel)], [t0.index], [t1.index])
                _put(order_matrix, [order], [t0.index], [t1.index])

                rel = get_relation(t1, t0, prefix=prefix)
                order = get_relation_value(rel, t1)

                _put(relation_type_matrix, [int(rel)], [t1.index], [t0.index])
                _put(order_matrix, [order], [t1.index], [t0.index])

    indices = list(range(len(d)))
    _put(relation_type_matrix, [int(RelationType.Equal)] * len(d), indices, indices)
    _put(order_matrix, [0] * len(d) , indices, indices)

    def build_mat(mat):
        assert len(set(zip(mat[1], mat[2]))) == len(list(zip(mat[1], mat[2])))
        return csr_matrix((mat[0], (mat[1], mat[2])), dtype=int)

    # distance_matrix = build_mat(distance_matrix)
    # order_matrix = build_mat(order_matrix)
    relation_type_matrix = build_mat(relation_type_matrix)
    order_matrix = build_mat(order_matrix)
    # 'distance': distance_matrix,
    # 'order': order_matrix,
    return {'relation': relation_type_matrix,
            'order': order_matrix}


MATRIX_BUILD = {
    # 'distance': _build_distance_matrix,
    'order': _build_distance_matrix,
    'relation': _build_distance_matrix,
    # 'ancestor': _build_ancestor_matrix
}



import copy
import itertools
import numpy as np

from ieml.syntax.terms import SyntaxTerm
from ..constants import LANGUAGES
from ..dictionary import Term, term
from ..syntax import Word
from .paths import path
from .tools import usl, replace_paths


class Template:
    def __init__(self, model, path_list):
        """

        :param model:
        :param path_list:
        :param translation_rule:
        """
        super().__init__()

        self.model = usl(model)

        if not isinstance(self.model.ieml_object, Word):
            raise ValueError("Template only implemented for words.")

        self.paths = [path(p) for p in path_list]
        if any(p.cardinal != 1 for p in self.paths):
            raise ValueError("Can only build a template from singular path (no '+')")

        self.multiples = []
        self.result = []

        self.build()

    def build(self):
        self.multiples = []
        for i, p in enumerate(self.paths):
            t = self.model[p]
            if not isinstance(t, SyntaxTerm) or t.script.cardinal == 1:
                raise ValueError("Invalid path for template creation [%s]->'%s' leading to a non SyntaxTerm object, "
                                 "or the term is not a paradigm."%(str(p), str(t)))

            self.multiples.append({
                'index': i,
                'path': p,
                'term': t,
            })

        self.template = np.zeros(shape=tuple(t['term'].script.cardinal for t in self.multiples), dtype=object)

        for index in itertools.product(*tuple(range(s) for s in self.template.shape)):
            self.template[index] = replace_paths(self.model, {
                m['path']: SyntaxTerm(term(m['term'].script.singular_sequences[index[i]])) for i, m in enumerate(self.multiples)
            })

    def get_translations(self, translation_template, mapping=None):
        """

        :param translation_template: str, a string with special string embedded to determine where to put the variation
        translations
        :param mapping: dict {path -> str} where str is the special string in translation template to replace with the
        term translation
        :return: dict {usl -> dict { str (language) ->  str}}
        """

        if mapping is None:
            mapping = {m['path']: str(m['path']) for m in self.multiples}

        def _replace(u):
            translation = copy.deepcopy(translation_template)
            for m in self.multiples:
                for l in LANGUAGES:
                    translation[l] = translation[l].replace(mapping[m['path']], u[m['path']].translations[l])
            return translation

        return {u: _replace(u) for u in self.__iter__()}


    def __iter__(self):
        return (np.asscalar(w) for w in np.nditer(self.template, flags=['refs_ok']))


import logging
import types
from functools import lru_cache
import os
import ply.yacc as yacc

from ieml.exceptions import InvalidScript
from ....exceptions import CannotParse
from ..script import AdditiveScript, MultiplicativeScript, NullScript
from ....constants import REMARKABLE_ADDITION
from ....commons import Singleton

from .lexer import get_script_lexer, tokens

from .... import parser_folder
import threading


class ScriptParser(metaclass=Singleton):
    tokens = tokens

    # ply have an internal state , then we forbid two thread try to parse a string simultaneously
    lock = threading.Lock()

    def __init__(self):
        self.t_add_rules()

        self.lexer = get_script_lexer()
        self.parser = yacc.yacc(module=self, errorlog=logging, start='term',
                                debug=False, optimize=True, picklefile=os.path.join(parser_folder, "script_parser.pickle"))
        # rename the parsing method (can't name it directly parse with lru_cache due to ply checking)
        self.parse = self.t_parse

    @lru_cache(maxsize=10000)
    def t_parse(self, s):
        with self.lock:
            try:
                return self.parser.parse(s, lexer=self.lexer)
            except InvalidScript as e:
                raise CannotParse(s, str(e))

    def p_error(self, p):
        if p:
            msg = "Syntax error at '%s' (%d, %d)" % (p.value, p.lineno, p.lexpos)
        else:
            msg = "Syntax error at EOF"

        raise InvalidScript(msg)

    # Rules
    def p_term(self, p):
        """ term : script_lvl_0
                | additive_script_lvl_0
                | script_lvl_1
                | additive_script_lvl_1
                | script_lvl_2
                | additive_script_lvl_2
                | script_lvl_3
                | additive_script_lvl_3
                | script_lvl_4
                | additive_script_lvl_4
                | script_lvl_5
                | additive_script_lvl_5
                | script_lvl_6
                | additive_script_lvl_6 """
        p[0] = p[1]

    def p_script_lvl_0(self, p):
        """ script_lvl_0 : PRIMITIVE LAYER0_MARK
                            | REMARKABLE_ADDITION LAYER0_MARK"""
        if p[1] == 'E':
            p[0] = NullScript(layer=0)
        elif p[1] in REMARKABLE_ADDITION:
            p[0] = AdditiveScript(character=p[1])
        else:
            p[0] = MultiplicativeScript(character=p[1])


    def p_additive_script_lvl_0(self, p):
        """ additive_script_lvl_0 : sum_lvl_0"""
        p[0] = AdditiveScript(children=p[1])

    def p_sum_lvl_0(self, p):
        """ sum_lvl_0 : script_lvl_0
                    | script_lvl_0 PLUS sum_lvl_0"""
        if len(p) == 4:
            p[3].append(p[1])
            p[0] = p[3]
        else:
            p[0] = [p[1]]

    def p_script_lvl_1(self, p):
        """ script_lvl_1 : additive_script_lvl_0 LAYER1_MARK
                        | additive_script_lvl_0 additive_script_lvl_0 LAYER1_MARK
                        | additive_script_lvl_0 additive_script_lvl_0 additive_script_lvl_0 LAYER1_MARK
                        | REMARKABLE_MULTIPLICATION LAYER1_MARK"""
        if isinstance(p[1], AdditiveScript):
            if len(p) == 3:
                p[0] = MultiplicativeScript(substance=p[1])
            elif len(p) == 4:
                p[0] = MultiplicativeScript(substance=p[1],
                                            attribute=p[2])
            else:
                p[0] = MultiplicativeScript(substance=p[1],
                                            attribute=p[2],
                                            mode=p[3])
        else:
            p[0] = MultiplicativeScript(character=p[1])

    def p_sum_lvl_1(self, p):
        """ sum_lvl_1 : script_lvl_1
                    |  script_lvl_1 PLUS sum_lvl_1"""
        if len(p) == 4:
            p[3].append(p[1])
            p[0] = p[3]
        else:
            p[0] = [p[1]]

    def p_additive_script_lvl_1(self, p):
        """ additive_script_lvl_1 : sum_lvl_1 """
        p[0] = AdditiveScript(children=p[1])

    def t_add_rules(self):
        def rule_layer(layer):
            def _sum(self, p):
                if len(p) == 4:
                    p[3].append(p[1])
                    p[0] = p[3]
                else:
                    p[0] = [p[1]]

            def _additive(self, p):
                p[0] = AdditiveScript(children=p[1])

            def _script(self, p):
                if len(p) == 3:
                    p[0] = MultiplicativeScript(substance=AdditiveScript(p[1]))
                elif len(p) == 4:
                    p[0] = MultiplicativeScript(substance=AdditiveScript(p[1]),
                                                attribute=AdditiveScript(p[2]))
                else:
                    p[0] = MultiplicativeScript(substance=AdditiveScript(p[1]),
                                                attribute=AdditiveScript(p[2]),
                                                mode=AdditiveScript(p[3]))

            result = []
            # sum rule
            function = types.FunctionType(_sum.__code__, _sum.__globals__)
            function.__name__ = 'p_sum_lvl_%d'%layer
            function.__doc__ = """sum_lvl_%d : script_lvl_%d
                                            | script_lvl_%d PLUS sum_lvl_%d"""%((layer,)*4)
            result.append(function)

            # additive parser rule
            function = types.FunctionType(_additive.__code__, _additive.__globals__)
            function.__name__ = 'p_additive_script_lvl_%d'%layer
            function.__doc__ = "additive_script_lvl_%d : sum_lvl_%d "%((layer,)*2)
            result.append(function)

            function = types.FunctionType(_script.__code__, _script.__globals__)
            function.__name__ = 'p_script_lvl_%d'%layer
            function.__doc__ = """script_lvl_%d : sum_lvl_%d LAYER%d_MARK
                                    | sum_lvl_%d sum_lvl_%d LAYER%d_MARK
                                    | sum_lvl_%d sum_lvl_%d sum_lvl_%d LAYER%d_MARK """%(
                (layer, layer - 1, layer) +
                (layer-1,)*2 + (layer,) +
                (layer-1,)*3 + (layer,))

            result.append(function)

            return result

        for i in range(2, 7):
            for f in rule_layer(i):
                setattr(self.__class__, f.__name__, f)


from .commons import IEMLSyntax
from ..constants import MAX_NODES_IN_SENTENCE
from ..exceptions import InvalidIEMLObjectArgument, InvalidTreeStructure
from .words import Word
from .tree_graph import TreeGraph


class AbstractClause(IEMLSyntax):
    def __init__(self, subtype, substance=None, attribute=None, mode=None, children=None):
        if not (substance or children):
            raise ValueError()

        if children is None:
            children = [substance, attribute, mode]
            if any(e is None for e in children):
                raise InvalidIEMLObjectArgument(self.__class__, "Must specify a substance, an attribute and a mode.")

        if len(children) != 3:
            raise InvalidIEMLObjectArgument(self.__class__, "A clause must have three children (%d provided)."%len(children))

        if not all(isinstance(e, subtype) for e in children):
            raise InvalidIEMLObjectArgument(self.__class__, "The children of a %s must be a %s instance."%
                                            (self.__class__.__name__, subtype.__name__))

        if children[0] == children[1]:
            raise InvalidIEMLObjectArgument(self.__class__, "The attribute and the substance (%s) must be distinct."%
                                            (str(children[0])))

        super().__init__(children)

    @property
    def substance(self):
        return self.children[0]

    @property
    def attribute(self):
        return self.children[1]

    @property
    def mode(self):
        return self.children[2]

    @property
    def grammatical_class(self):
        return self.attribute.grammatical_class

    def compute_str(self, children_str):
        return '('+'*'.join(children_str)+')'


class AbstractSentence(IEMLSyntax):
    closable = True

    def __init__(self, subtype, children, literals=None):
        try:
            _children = tuple(e for e in children)
        except TypeError:
            raise InvalidIEMLObjectArgument(self.__class__, "The argument %s is not an iterable" % str(children))

        if not all(isinstance(e, subtype) for e in _children):
            raise InvalidIEMLObjectArgument(self.__class__, "The children of a %s must be %s instance."%
                                            (self.__class__.__name__, subtype.__name__))

        try:
            self.tree_graph = TreeGraph(_children)
        except InvalidTreeStructure as e:
            raise InvalidIEMLObjectArgument(self.__class__, e)

        if len(self.tree_graph.nodes) > MAX_NODES_IN_SENTENCE:
            raise InvalidIEMLObjectArgument(self.__class__, "Too many distinct nodes: %d>%d."%
                                            (len(self.tree_graph.nodes), MAX_NODES_IN_SENTENCE))

        super().__init__((e for stage in self.tree_graph.stages
                          for e in sorted((t[1] for s in stage for t in self.tree_graph.transitions[s]))),
                         literals=literals)

    @property
    def grammatical_class(self):
        return self.tree_graph.root.grammatical_class

    def compute_str(self, children_str):
        return '[' + '+'.join(children_str) + ']'


class Clause(AbstractClause):
    def __init__(self, substance=None, attribute=None, mode=None, children=None):
        super().__init__(Word, substance=substance, attribute=attribute, mode=mode, children=children)


class Sentence(AbstractSentence):
    def __init__(self, children, literals=None):
        super().__init__(Clause, children=children, literals=literals)


class SuperClause(AbstractClause):
    def __init__(self, substance=None, attribute=None, mode=None, children=None):
        super().__init__(Sentence, substance=substance, attribute=attribute, mode=mode, children=children)


class SuperSentence(AbstractSentence):
    def __init__(self, children, literals=None):
        super().__init__(SuperClause, children=children, literals=literals)



from collections import defaultdict
import numpy

from ..exceptions import InvalidTreeStructure


class TreeGraph:
    def __init__(self, list_transitions):
        """
        Transitions list must be the (start, end, data) the data will be stored as the transition tag
        :param list_transitions:
        """
        # transitions : dict
        #
        self.transitions = defaultdict(list)
        for t in list_transitions:
            self.transitions[t[0]].append((t[1], t))

        self.nodes = sorted(set(self.transitions) | {e[0] for l in self.transitions.values() for e in l})

        # sort the transitions
        for s in self.transitions:
            self.transitions[s].sort(key=lambda t: self.nodes.index(t[0]))

        self.nodes_index = {n: i for i, n in enumerate(self.nodes)}
        _count = len(self.nodes)
        self.array = numpy.zeros((len(self.nodes), len(self.nodes)), dtype=bool)

        for t in self.transitions:
            for end in self.transitions[t]:
                self.array[self.nodes_index[t]][self.nodes_index[end[0]]] = True

        # checking
        # root checking, no_parent hold True for each index where the node has no parent
        parents_count = numpy.dot(self.array.transpose().astype(dtype=int), numpy.ones((_count,), dtype=int))
        no_parents = parents_count == 0
        roots_count = numpy.count_nonzero(no_parents)

        if roots_count == 0:
            raise InvalidTreeStructure('No root node found, the graph has at least a cycle.')
        elif roots_count > 1:
            raise InvalidTreeStructure('Several root nodes found.')

        self.root = self.nodes[no_parents.nonzero()[0][0]]

        if (parents_count > 1).any():
            raise InvalidTreeStructure('A node has several parents.')

        def __stage():
            current = [self.root]
            while current:
                yield current
                current = [child[0] for parent in current for child in self.transitions[parent]]

        self.stages = list(__stage())


import ply.lex as lxr
import logging

logger = logging.getLogger(__name__)

TERM_REGEX = r'[EUASBTOMFIacbedgfihkjmlonpsutwyx][EUASBTOMFIacbedgfihkjmlonpsutwyx\.\-\;\:\,\'\’\_\+]+'

tokens = (
   'TERM',
   'PLUS',
   'TIMES',
   'LPAREN',
   'RPAREN',
   'LBRACKET',
   'RBRACKET',
   'L_CURLY_BRACKET',
   'R_CURLY_BRACKET',
   'SLASH',
   'LITERAL',
)


def get_lexer(module=None):
    t_TERM = TERM_REGEX
    t_PLUS   = r'\+'
    t_TIMES   = r'\*'
    t_LPAREN  = r'\('
    t_RPAREN  = r'\)'
    t_LBRACKET = r'\['
    t_RBRACKET  = r'\]'
    t_L_CURLY_BRACKET = r'\{'
    t_R_CURLY_BRACKET = r'\}'
    t_SLASH = r'\/'
    t_LITERAL = r'\<(\\\>|[^\>])+\>'
#    t_USL_TAG = r'([A-Za-z0-9 _\./\\-]+)'

    t_ignore  = ' \t\n'

    # Error handling rule
    def t_error(t):
        logger.log(logging.ERROR, "Illegal character '%s'" % t.value[0])
        t.lexer.skip(1)

    return lxr.lex(module=module, errorlog=logging)



from itertools import islice

from ieml.syntax.terms import SyntaxTerm
from ..constants import LANGUAGES
from ..dictionary import Term
from .paths import Path, enumerate_paths, resolve


class Usl:
    def __init__(self, ieml_object):
        self.ieml_object = ieml_object
        self._rules = None
        self._level = {}

    def __str__(self):
        return str(self.ieml_object)

    def __eq__(self, other):
        if isinstance(other, Usl):
            return self.ieml_object.__eq__(other.ieml_object)

    def __hash__(self):
        return hash(self.ieml_object)

    @property
    def paths(self):
        return self.rules(SyntaxTerm)

    def __getitem__(self, item):
        if isinstance(item, Path):
            if item in self.paths:
                return self.paths[item]

            return resolve(self.ieml_object, item)

        raise NotImplemented

    def auto_translation(self):
        result = {}
        entries = sorted([t for p, t in self.paths.items()])
        for l in LANGUAGES:
            result[l] = ' '.join((e.translations[l] for e in islice(entries, 10)))

        return result

    def rules(self, type):
        if type not in self._level:
            self._level[type] = {path: element for path, element in enumerate_paths(self.ieml_object, level=type)}

        return self._level[type]

    def objects(self, type):
        return set(self.rules(type).values())


from collections import defaultdict

import numpy

from ieml.exceptions import InvalidPathException
from ieml.syntax.terms import SyntaxTerm
from ieml.tools import ieml
from ...exceptions import InvalidIEMLObjectArgument
from ...syntax import SuperSentence, Sentence, Clause, SuperClause, Text, Word, Morpheme
from ...dictionary import Term
from .exceptions import IEMLObjectResolutionError, ResolveError
from .parser import PathParser
from .paths import Path, Coordinate, MultiplicativePath, AdditivePath, ContextPath


def path(p):
    if isinstance(p, Path):
        return p
    if isinstance(p, str):
        return PathParser().parse(p)

    try:
        l = list(p)
        if all(isinstance(e, Path) for e in l):
            res = l[0]
            for p in l[1:]:
                res *= p
            return res

    except TypeError:
        pass

    raise ValueError("Invalid argument to create a path.")


def _resolve_path_tree_graph(tree_graph, path):
    if isinstance(path, Coordinate):
        coords = [path]
    else:
        coords = list(path.children)

    c0 = coords[0]
    if c0.kind == 's':
        stack = {tree_graph.root}
    elif c0.kind == 'a':
        # tous les attributs
        stack = {attr[0] for vals in tree_graph.transitions.values() for attr in vals}
    else:
        # tous les modes
        return {attr[1][2] for vals in tree_graph.transitions.values() for attr in vals}

    for c in coords[1:]:
        _stack = set()
        for s in stack:
            if c.kind == 's':
                raise InvalidPathException(tree_graph, path, "double substance 's' (root node).")

            if c.kind == 'm':
                if c.index is not None:
                    _stack.add(tree_graph.transitions[s][c.index][1][2])
                else:
                    _stack |= {vals[1][2] for vals in tree_graph.transitions[s]}

            if c.kind == 'a':
                if c.index is not None:
                    _stack.add(tree_graph.transitions[s][c.index][0])
                else:
                    _stack |= {vals[0] for vals in tree_graph.transitions[s]}

        if not _stack:
            return set()

        stack = _stack

    return stack


def _resolve_path(obj, path):
    """path is a mul of coord or a coord"""
    if obj.__class__ not in path.context.accept:
        return set()

    if isinstance(obj, Text):
        if path.index:
            return {obj.children[path.index]}

        return set(obj.children)

    if isinstance(obj, (Sentence, SuperSentence)):
        return _resolve_path_tree_graph(obj.tree_graph, path)

    if isinstance(obj, Word):
        if path.kind == 'r':
            if path.index is not None:
                return {obj.root.children[path.index]}
            return {obj.root}
        else:
            if path.index is not None:
                return {obj.flexing.children[path.index]}
            if obj.flexing:
                return {obj.flexing}
            else:
                raise InvalidPathException(obj, path, "no flexing morpheme in this Word.")

    if isinstance(obj, Term):
        raise InvalidPathException(obj, path, "can't deference a Term.")


def resolve(ieml_object, path):
    if ieml_object.__class__ not in path.context.accept:
        raise InvalidPathException(ieml_object, path, "invalid path context, expected types: {%s}."%', '.join(path.context.accept))

    result = set()
    for d in path.develop:
        if isinstance(d, (Coordinate, MultiplicativePath)):
            result |= _resolve_path(ieml_object, d)
        else:
            # context path
            stack = {ieml_object}
            for c in d.children:
                _stack = set()
                for s in stack:
                    _stack |= _resolve_path(s, c)

                if not _stack:
                    break

                stack = _stack
            else:
                result |= stack

    return result


def _enumerate_paths(ieml_obj, level):
    if isinstance(ieml_obj, level):
        yield [], ieml_obj

    if isinstance(ieml_obj, Text):
        for i, t in enumerate(ieml_obj.children):
            for p, e in _enumerate_paths(t, level=level):
                yield [path('t%d'%i)] + p, e

    if isinstance(ieml_obj, (Sentence, SuperSentence)):
        for node in set(node for clause in ieml_obj for node in clause):
            for p, e in _enumerate_paths(node, level=level):
                yield [_tree_graph_path_of_node(ieml_obj.tree_graph, node)] + p, e

    if isinstance(ieml_obj, Word):
        for i, t in enumerate(ieml_obj.root.children):
            for p, e in _enumerate_paths(t, level=level):
                yield [path('r%d'%i)] + p, e

        if ieml_obj.flexing:
            for i, t in enumerate(ieml_obj.flexing.children):
                for p, e in _enumerate_paths(t, level=level):
                    yield [path('f%d' % i)] + p, e

    return


def _tree_graph_path_of_node(tree_graph, node):
    if node in tree_graph.nodes:
        nodes = [(node, False)]
    else:
        nodes = []

    # can be a mode
    nodes += [(c[0], True) for c_list in tree_graph.transitions.values() for c in c_list if c[1][2] == node]
    if not nodes:
        raise ValueError("Node not in tree graph : %s" % str(node))

    def _build_coord(node, mode=False):
        if node == tree_graph.root:
            return [Coordinate(kind='s')]

        parent = tree_graph.nodes[numpy.where(tree_graph.array[:, tree_graph.nodes_index[node]])[0][0]]

        return _build_coord(parent) + \
               [Coordinate(index=[c[0] for c in tree_graph.transitions[parent]].index(node), kind='m' if mode else 'a')]

    return AdditivePath([MultiplicativePath(_build_coord(node, mode)) for node, mode in nodes])


def enumerate_paths(ieml_obj, level=SyntaxTerm):
    for p, t in _enumerate_paths(ieml_obj, level=level):
        if len(p) == 1:
            yield p[0], t
        else:
            yield ContextPath(p), t


def _build_deps_tree_graph(rules):
    def _node():
        return {
            'resolve': defaultdict(_node),  # the indexed coordinate
            'context': [],
            'rules': defaultdict(_node)
        }

    # contain s, a et m (relative path to the root)
    roots = defaultdict(_node)

    s = Coordinate(kind='s')

    for p, e in rules:
        if isinstance(p, ContextPath):
            actual_P = p.children[0]  # multiplicative or coordinate
            ctx_P = path(p.children[1:])
        else:
            actual_P = p
            ctx_P = None

        if isinstance(actual_P, MultiplicativePath):
            actual_P = list(actual_P.children)
        else:
            actual_P = [actual_P]

        # actual_p is the list of coordinate to navigate
        # ctx_p is the rest of the path or None

        current_node = roots

        # replacing s0 -> s
        if actual_P[0].kind == 's' and actual_P[0].index == 0:
            actual_P[0] = s

        first = True
        for c in actual_P:
            if c.index is not None:
                categorie = 'resolve'
            else:
                categorie = 'rules'

            if first:
                current_node = current_node[c]
                first = False
            else:
                current_node = current_node[categorie][c]

        # the nodes with ctx_P is None will be instanciate
        current_node['context'].append((ctx_P, e))

    result = []

    def _merge_nodes(n0, n1):
        for cat in ('resolve', 'rules'):
            for r in n1[cat]:
                if r in n0[cat]:
                    _merge_nodes(n0[cat][r], n1[cat][r])
                else:
                    n0[cat][r] = n1[cat][r]

        # merge ctx
        n0['context'] += n1['context']

    coord_a = Coordinate('a')
    coord_m = Coordinate('m')

    # apply the rules on the tree starting from s
    def _apply_rule(node, _path, mode=False):
        # we add the globals rules to its own
        _merge_nodes(node['rules'][coord_a], roots[coord_a])
        _merge_nodes(node['rules'][coord_m], roots[coord_m])

        obj = _resolve_ctx(_path, node['context'])
        error = obj is None

        if mode or not node['resolve']:
            return error, obj

        # resolve the children
        max_i = max(n.index for n in node['resolve'])

        _result = {
            'a': [],
            'm': []
        }

        for r in ('a', 'm'):
            indexed = {k.index: node['resolve'][k] for k in node['resolve'] if k.kind == r}
            rules = [k for k in node['rules'] if k.kind == r]

            if max(indexed, default=-1) + 1 > len(indexed) + (1 if rules else 0):
                raise ResolveError("Not enough information to instantiate all the %s rules. At %s."%(r, _path))

            for i in range(max_i + 1):
                if i in indexed:
                    child = indexed[i]
                else:
                    child = _node()

                _merge_nodes(child, node['rules'][r])

                e, n = _apply_rule(child, _path + "%s%d"%(r, i), mode=r == 'm')
                error |= e
                _result[r].append(n)

        for a, m in zip(_result['a'], _result['m']):
            result.append((obj, a, m))

        return error, obj
    # 'context' attribute is now the sub element
    # generating the triplet from the root s

    error, o = _apply_rule(roots[s], 's0')

    return error, result[::-1]


def _build_deps_text(rules):
    indexes = defaultdict(list)
    generals = []

    for p, e in rules:
        if isinstance(p, ContextPath):
            actual_P = p.children[0]  # coordinate
            ctx_P = path(p.children[1:])
        else:
            actual_P = p
            ctx_P = None

        if not isinstance(actual_P, Coordinate):
            raise ResolveError("A text must be defined by a coordinate, not %s"%actual_P)

        if actual_P.index is not None:
            indexes[actual_P.index].append((ctx_P, e))
        else:
            generals.append((ctx_P, e))

    if not indexes:
        return False, []

    if max(indexes, default=-1) + 1 - len(indexes) > 1:
        # if there is more than one missing
        raise ResolveError("Index missing on text definition.")

    error = False
    i = 0
    result = []
    while i <= max(indexes):
        ctx_rules = generals[:]
        if i in indexes:
            ctx_rules.extend(indexes[i])

        node = _resolve_ctx("t%d"%i, ctx_rules)
        error |= node is None
        result.append(node)
        i += 1

    return error, result


def _build_deps_word(rules):
    result = {
        'r': {
            'indexes': defaultdict(list),
            'generals': []
        },
        'f': {
            'indexes': defaultdict(list),
            'generals': []
        }
    }

    for p, e in rules:
        if isinstance(p, ContextPath):
            # should be an error but support it
            actual_P = p.children[0]  # coordinate
            ctx_P = path(p.children[1:])
        else:
            actual_P = p
            ctx_P = None

        if not isinstance(actual_P, Coordinate):
            raise ResolveError("A word must be defined by a single coordinate.")

        if actual_P.index is not None:
            result[actual_P.kind]['indexes'][actual_P.index].append((ctx_P, e))
        else:
            result[actual_P.kind]['generals'].append((ctx_P, e))

    error = False
    for k in result:
        # k for kind (r or f)

        indexes = result[k]['indexes']
        generals = result[k]['generals']

        if not indexes and not generals:
            result[k] = []
            continue

        if max(indexes, default=-1) + 1 > len(indexes) + len(generals):
            # if there is more than one missing
            raise ResolveError("Index missing on word definition.")

        current = []
        length = len(indexes) + len(generals)
        generals = generals.__iter__()
        for i in range(length):
            if i in indexes:
                ctx_rules = indexes[i]
            else:
                ctx_rules = [next(generals)]

            node = _resolve_ctx("%s%d"%(k, i), ctx_rules)
            error |= node is None

            current.append(node)

        result[k] = current

    return error, result


def _inferred_types(path, e):
    result = set()
    for inf in path.context.switch:
        if isinstance(e, tuple(path.context.switch[inf])):
            result.add(inf)

    if result:
        return result

    raise ResolveError("No compatible type found with the path %s and the ieml object of type %s"%
                     (str(path), e.__class__.__name__))


_errors = []
_context_stack = []
debug = False


def _context_error_handler(func):
    def wrapper(_path, rules):
        _context_stack.append(_path)

        result = None
        if debug:
            result = func(rules)
        else:
            try:
                result = func(rules)
            except ResolveError as e:
                _errors.append((':'.join(_context_stack[1:]), str(e)))
            except InvalidIEMLObjectArgument as e:
                _errors.append((':'.join(_context_stack[1:]), str(e)))

        _context_stack.pop()

        return result

    return wrapper


@_context_error_handler
def _resolve_ctx(rules):
    """
    Resolve the context of the rules (the type of this element), and building the ieml element.
    :param rules:
    :return:
    """
    if not rules:
        raise ResolveError("Missing node definition.")

    # if rules == [(None, e)] --> e
    if len(rules) == 1 and rules[0][0] is None:
        return rules[0][1]

    if any(r[0] is None for r in rules):
        raise ResolveError("Multiple definition, multiple ieml object provided for the same node.")

    if any(not isinstance(r[0], Path) for r in rules):
        raise ResolveError("Must have only path instance.")

    # resolve all the possible types for this element
    r0 = rules[0]
    types = _inferred_types(*r0)
    for r in rules[1:]:
        types = types.intersection(_inferred_types(*r))

    if not types:
        raise ResolveError("No definition, no type inferred on rules list.")

    if len(types) > 1:
        raise ResolveError("Multiple definition, multiple type inferred on rules list.")

    type = next(types.__iter__())

    if type == Word:
        error, deps = _build_deps_word(rules)
        if error:
            return

        flexing = None
        if deps['f']:
            flexing = Morpheme(deps['f'])
        if not deps['r']:
            raise ResolveError("No root for the word node.")

        return Word(Morpheme(deps['r']), flexing)

    if type == Text:
        error, deps = _build_deps_text(rules)
        if error:
            return

        return Text(deps)

    if type in (SuperSentence, Sentence):
        error, deps = _build_deps_tree_graph(rules)
        if error:
            return

        if type == Sentence:
            clauses = []
            for s, a, m in deps:
                clauses.append(Clause(s, a, m))
            return Sentence(clauses)
        else:
            clauses = []
            for s, a, m in deps:
                clauses.append(SuperClause(s, a, m))
            return SuperSentence(clauses)

    raise ResolveError("Invalid type inferred %s"%type.__name__)


def resolve_ieml_object(paths, elements=None):
    global _errors, _context_stack
    _errors = []
    _context_stack = []

    if isinstance(paths, dict):
        paths = list(paths.items())

    if elements is None:
        result = _resolve_ctx('', [(d, ieml(e)) for p, e in paths for d in path(p).develop])
    else:
        result = _resolve_ctx('', [(d, ieml(e)) for e, p in zip(elements, paths) for d in path(p).develop])

    if _errors:
        raise IEMLObjectResolutionError(_errors)

    return result




import logging
from collections import namedtuple

from ..commons import cached_property
from ..constants import LANGUAGES
from .script import script as _script
logger = logging.getLogger(__name__)
Translations = namedtuple('Translations', list(LANGUAGES))
Translations.__getitem__ = lambda self, item: self.__getattribute__(item) if item in LANGUAGES \
    else tuple.__getitem__(self, item)


class Term:
    # closable = True

    def __init__(self, script, index, dictionary, parent):
        super().__init__()

        self.dictionary = dictionary
        self.parent = parent

        self.script = _script(script)

        self.index = index

        # if term in a dictionary, those values will be set
        self.translations = Translations(**{l: self.dictionary.translations[l][self.script] for l in LANGUAGES})

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.index == other.index

    def __gt__(self, other):
        return self.index > other.index

    def __hash__(self):
        return hash(self.__str__())

    def __str__(self):
        return "[%s]" % str(self.script)

    _compute_str = __str__

    def tree_iter(self):
        yield self

    @cached_property
    def root(self):
        if self.parent is None:
            return self
        else:
            return self.parent.root

    @property
    def is_root(self):
        return self in self.dictionary.roots

    @property
    def inhibitions(self):
        return self.dictionary.inhibitions[self.root]

    @cached_property
    def max_rank(self):
        if self.parent is None:
            # root
            return max(t.rank for t in self.relations.contains if len(t) != 1)

        return self.root.max_rank

    @property
    def empty(self):
        return self.script.empty

    @cached_property
    def ntable(self):
        return sum(self.script.cells[i].shape[2] for i in range(len(self.script.cells)))

    @cached_property
    def tables_term(self):
        return [self.dictionary.terms[s] for s in self.script.tables_script]

    @property
    def grammatical_class(self):
        return self.script.script_class

    @cached_property
    def singular_sequences(self):
        return [self.dictionary.terms[ss] for ss in self.script.singular_sequences]

    @property
    def layer(self):
        return self.script.layer

    @cached_property
    def table(self):
        return self.dictionary.tables[self.root][self]

    @cached_property
    def relations(self):
        return self.dictionary.relations_graph[self]

    def __contains__(self, item):
        from .tools import term
        if not isinstance(item, Term):
            item = term(item, dictionary=self.dictionary)
        elif item.dictionary != self.dictionary:
            logger.log(logging.ERROR, "Comparison between different dictionary.")
            return False

        return item.script in self.script

    def __len__(self):
        return self.script.cardinal

    def __iter__(self):
        return self.singular_sequences.__iter__()
